/**
  *  \file client/dialogs/processmgr.cc
  *  \brief Process Manager
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/processmgr.h"
#include "client/widgets/decaymsg.h"
#include "client/widgets/helpwidget.h"
#include "client/widgets/processlist.h"
#include "cpluslib/holder.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "int/exec.h"
#include "int/if/userif.h"
#include "int/process.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "int/notify.h"
#include "client/dialogs/notifydlg.h"

namespace {
    class WProcessManagerDialog : public UIWindow,
                                  public IntUserInterfaceBinding
    {
        ObjectHolder h;

        WProcessList* process_list;
        UIButton* btn_g;

        void init();
        void setProcessState(IntExecutionContext::ProcessState state, bool all);
        void changePriority();
        void performChanges();
     public:
        WProcessManagerDialog();
        bool handleEvent(const UIEvent& event, bool second_pass);
        void findObject(GObject* obj);

        // IntUserInterfaceBinding
        int  getProperty(IntUserInterfaceProperty p);
        void doGoToScreen(int screen, int obj);
        void doGoToChart(GPoint pt);
        void doMoveScanner(GPoint pt);
        void doPopupConsole();
        void doOverlayMessage(string_t msg);
        void doLeaveRace();
        void doSetChartView(string_t name);
        IntKeymapRef getKeymap();
    };
}

/** Change a process' state. This checks that only processes are
    affected which the user is allowed to modify.
    \param proc The process
    \param state New state */
static bool
changeProcessState(IntExecutionContext& proc, IntExecutionContext::ProcessState state)
{
    switch (proc.getState()) {
     case IntExecutionContext::psTerminating:
     case IntExecutionContext::psSuspended:
     case IntExecutionContext::psRunnable:
        proc.setState(state);
        return true;
     default:
        return false;
    }
}


/** Constructor. */
WProcessManagerDialog::WProcessManagerDialog()
    : UIWindow(_("Process Manager"), BLUE_WINDOW, COLOR_GRAY)
{
    init();
}

/** Build the dialog. */
void
WProcessManagerDialog::init()
{
    // VBox
    //   VBox [instance0]
    //     WProcessListHeader
    //     UIFrameGroup [HBox instance0]
    //       WProcessList
    //       UIScrollbar
    //   HBox
    //     UIButton "R"
    //     UIStaticText "Run"
    //     UIButton "T"
    //     UIStaticText "Terminate"
    //     UIButton "S"
    //     UIStaticText "Suspend"
    //    [UIButton "L"
    //     UIStaticText "Locks..."]
    //     UISpacer
    //     UIStaticText "Priority"
    //     UIButton "P"
    //   HBox
    //     UIButton "X-Execute"
    //     UIButton "G-Go to"
    //     UISpacer
    //     UIButton "Close"
    //     UIButton "Help"

    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1       = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIFrameGroup& g11 = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    UIGroup& g2       = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g3       = h.add(new UIGroup(UIHBoxLayout::instance5));

    g1.add(h.add(new WProcessListHeader()));
    g1.add(g11);

    process_list = &h.add(new WProcessList());
    process_list->update();
    g11.add(*process_list);
    g11.add(h.add(new UIScrollbar(*process_list)));

    UIButton* btn = &h.add(new UIButton("R", 'r'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Run"), UIStaticText::StaticNormal)));

    btn = &h.add(new UIButton("T", 't'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Terminate"), UIStaticText::StaticNormal)));

    btn = &h.add(new UIButton("S", 's'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(*btn);
    g2.add(h.add(new UIStaticText(_("Suspend"), UIStaticText::StaticNormal)));

    g2.add(h.add(new UISpacer()));

    btn = &h.add(new UIButton("N", 'n'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(h.add(new UIStaticText(_("Notification"), UIStaticText::StaticNormal)));
    g2.add(*btn);

    // btn = &h.add(new UIButton("L", 'l'));
    // btn->setState(UIButton::bf_SmallSize, true);
    // g2.add(h.add(new UIStaticText(_("Locks"), UIStaticText::StaticNormal)));
    // g2.add(*btn);

    btn = &h.add(new UIButton("P", 'p'));
    btn->setState(UIButton::bf_SmallSize, true);
    g2.add(h.add(new UIStaticText(_("Priority"), UIStaticText::StaticNormal)));
    g2.add(*btn);

    btn_g = &h.add(new UIButton(_("G - Go To"), 'g'));
    g3.add(h.add(new UIButton(_("X - Execute"), 'x')));
    g3.add(*btn_g);
    g3.add(h.add(new UISpacer()));
    g3.add(h.add(new UIButton(1000, _("Close"), SDLK_ESCAPE)).addStop());
    g3.add(h.add(new UIButton(_("Help"), 'h')));

    add(g1);
    add(g2);
    add(g3);
    add(h.add(new WHelpWidget("pcc2:processmgr")));

    process_list->setFocusedWidget();

    pack();
}

/** Set process state. Handles user interface interaction.
    \param state Target state
    \param all true to affect all processes, false to only affect current */
void
WProcessManagerDialog::setProcessState(IntExecutionContext::ProcessState state, bool all)
{
    if (all) {
        bool change = false;
        const IntProcessList& pl = getProcessList();
        for (uint32_t i = 0; i < pl.size(); ++i) {
            if (changeProcessState(*pl[i], state)) {
                change = true;
            }
        }
        if (change) {
            process_list->drawWidget();
        }
    } else {
        IntExecutionContext* proc = process_list->getSelectedProcess();
        if (proc && changeProcessState(*proc, state)) {
            process_list->updateCurrent();
        }
    }
}

/** Change process priority. This allows the user to change the
    current current process' priority. */
void
WProcessManagerDialog::changePriority()
{
    if (IntExecutionContext* proc = process_list->getSelectedProcess()) {
        UIDecimalSelector sel(0, 0, 99, 10);
        sel.setValue(proc->getPriority());
        if (sel.doStandardDialog(_("Process Manager"), _("Enter new process priority:"), 0)) {
            /* Change priority */
            proc->setPriority(sel.getValue());
            handlePriorityChange(*proc);

            /* Update cursor */
            const IntProcessList& pl = getProcessList();
            for (uint32_t i = 0; i < pl.size(); ++i) {
                if (pl[i] == proc) {
                    process_list->scrollTo(i, 1);
                    break;
                }
            }
            process_list->drawWidget();
        }
    }
}

/** Perform pending T/R/S requests. */
void
WProcessManagerDialog::performChanges()
{
    runRunnableProcesses();
    killTerminatedProcesses();
    process_list->update();
}

bool
WProcessManagerDialog::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case 't':
         case 't' + ss_Ctrl:
            setProcessState(IntExecutionContext::psTerminating, (event.key.code & ss_Ctrl) != 0);
            return true;
         case 's':
         case 's' + ss_Ctrl:
            setProcessState(IntExecutionContext::psSuspended, (event.key.code & ss_Ctrl) != 0);
            return true;
         case 'r':
         case 'r' + ss_Ctrl:
            setProcessState(IntExecutionContext::psRunnable, (event.key.code & ss_Ctrl) != 0);
            return true;
         case 'p':
            changePriority();
            return true;
         case 'x':
            performChanges();
            return true;
         case 'g':
            if (IntExecutionContext* proc = process_list->getSelectedProcess()) {
                GObject* obj = proc->getInvokingObject();
                if (dynamic_cast<GPlanet*>(obj)) {
                    if (proc->getProcessKind() == IntExecutionContext::pkBaseTask) {
                        doGoToScreen(3, obj->getId());
                    } else {
                        doGoToScreen(2, obj->getId());
                    }
                } else if (dynamic_cast<GShip*>(obj)) {
                    doGoToScreen(1, obj->getId());
                }
            }
            return true;
         case 'n':
            if (IntExecutionContext* proc = process_list->getSelectedProcess()) {
                if (IntNotificationMessage* msg = proc->getNotificationMessage()) {
                    // Move to that message
                    IntNotificationMessageStore& inms = IntNotificationMessageStore::instance;
                    inms.setCurrent(inms.getIndexOfMessage(msg));

                    // Suspend the process, so that the message does not die in our hands
                    setProcessState(IntExecutionContext::psSuspended, false);

                    // Tell caller to show messages
                    stop(1001);
                }
            }
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
WProcessManagerDialog::findObject(GObject* obj)
{
    const IntProcessList& pl = getProcessList();
    for (uint32_t i = 0; i < pl.size(); ++i) {
        if (pl[i]->getInvokingObject() == obj) {
            process_list->scrollTo(i, 1);
            break;
        }
    }
}

int
WProcessManagerDialog::getProperty(IntUserInterfaceProperty p)
{
    return getPreviousBinding()->getProperty(p);
}

void
WProcessManagerDialog::doGoToScreen(int screen, int obj)
{
    stop(1);
    getPreviousBinding()->doGoToScreen(screen, obj);
}

void
WProcessManagerDialog::doGoToChart(GPoint pt)
{
    stop(1);
    getPreviousBinding()->doGoToChart(pt);
}

void
WProcessManagerDialog::doMoveScanner(GPoint pt)
{
    stop(1);
    getPreviousBinding()->doMoveScanner(pt);
}

void
WProcessManagerDialog::doPopupConsole()
{
    stop(1);
    getPreviousBinding()->doPopupConsole();
}

void
WProcessManagerDialog::doOverlayMessage(string_t msg)
{
    showDecayingMessage(msg);
}

void
WProcessManagerDialog::doLeaveRace()
{
    stop(1);
    getPreviousBinding()->doLeaveRace();
}

void
WProcessManagerDialog::doSetChartView(string_t name)
{
    stop(1);
    getPreviousBinding()->doSetChartView(name);
}

IntKeymapRef
WProcessManagerDialog::getKeymap()
{
    return getPreviousBinding()->getKeymap();
}

void
doProcessManager(GObject* invoker)
{
    int result;
    {
        WProcessManagerDialog dlg;
        if (invoker) {
            dlg.findObject(invoker);
        }
        ui_root->centerWidget(dlg);
        result = ui_root->runModal(dlg);
    }

    if (result == 1000 || result == 1001) {
        /* This means the dialog was closed with ESC. Perform actions.
           This must be done when the WProcessManagerDialog is already
           destroyed, so it uses just the parent IntUserInterfaceBinding.
           If the dialog was closed otherwise, the parent
           IntUserInterfaceBinding will arrange for processes to run. */
        runRunnableProcesses();
        killTerminatedProcesses();
    }
    if (result == 1001) {
        /* Notification messages */
        doNotifyScreen();
    }
}
