/**
  *  \file client/screens/planettaskscreen.cc
  *  \brief Planet Task Screen
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/planettaskscreen.h"
#include "cpluslib/format.h"
#include "game/global.h"
#include "game/obj.h"
#include "game/planet.h"
#include "int/exec.h"
#include "int/if/planetif.h"
#include "int/intglobal.h"
#include "int/process.h"
#include "util/nls.h"
#include "client/wscanchart.h"
#include "client/tasksupport.h"
#include "util/strlist.h"
#include "ui/listbox.h"
#include "ui/util.h"
#include "client/dlg-fcode.h"
#include "game/fcode.h"
#include "int/if/basepred.h"

namespace {
    class WPlanetAutoTaskScreen : public WAutoTaskScreen {
     public:
        WPlanetAutoTaskScreen(WPlanetAutoTaskType& type);
        void update();
        void setScriptContext();
    };
}

/************************** WPlanetAutoTaskType **************************/

/** Constructor.
    \param base Underlying real object type
    \param pk Process kind to use for locating auto tasks */
WPlanetAutoTaskType::WPlanetAutoTaskType(GObjectType& base, IntExecutionContext::ProcessKind pk)
    : WAutoTaskObjectType(base),
      pk(pk)
{ }

IntExecutionContext*
WPlanetAutoTaskType::getTask(bool create)
{
    return getAutoTaskForObject(getObject(getCurrent()), pk, create);
}

/** Create a "SetFCode" command. Asks the user for the fcode. */
void
WPlanetAutoTaskType::createSetFCodeCommand()
{
    // Find out old friendly code. We don't have any predictor
    // specifically for planets, so use the base predictor.
    GPlanet* pl = dynamic_cast<GPlanet*>(&getObject(getCurrent()));
    if (!pl) {
        return;
    }
    WBaseTaskPredictor pred(*pl);
    pred.predictTask(getEditor(), getCursor());

    // Dialog
    GFCodeList list(GFCodeList::getInstance(), *pl);
    WFCodeWindow dialog(_("Change Friendly Code"), list);
    dialog.setFCode(pred.getPlanet().getFCode());
    ui_root->centerWidget(dialog);
    if (ui_root->runModal(dialog) == cm_Enter && !dialog.getFCode().empty()) {
        insert("SetFCode " + quoteString(dialog.getFCode()));
    }
}

/** Create a cargo transfer command. This is used for the "2-Cargo" button
    and will offer a menu.
    \param sender Invoking button as anchor for menu */
void
WPlanetAutoTaskType::createCargoCommand(UIBaseWidget& sender)
{
    // This is very rudimentary. PCC 1.x has nothing at all.
    StringList sl;
    sl.add(1, _("Unload All Ships"));
    UIStandardListbox list(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    if (list.doStandardMenu(sender)) {
        switch (list.getKey()) {
         case 1:
            insert("CargoUnloadAllShips");
            break;
        }
    }
}

/************************* WPlanetAutoTaskScreen *************************/

WPlanetAutoTaskScreen::WPlanetAutoTaskScreen(WPlanetAutoTaskType& type)
    : WAutoTaskScreen(type, WScreenId::PlanetTaskScreen)
{
    createTileLayout(*this, "PLANETTASKSCREEN", type);
    addChart(holder.add(new WScannerChartWidget()));
    setKeymap(IntKeymap::getKeymapByName("PLANETTASKSCREEN"));
}

void
WPlanetAutoTaskScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(&getType().getObject(getType().getCurrent()))) {
        WScannerChartWidget& chart = *getScanner().getChart();
        chart.setCenterPosition(mo->getPos());
        chart.setScannerOrigin(mo->getPos());
        chart.setScannerTarget(mo->getPos());
    }
}

void
WPlanetAutoTaskScreen::setScriptContext()
{
    global_contexts.push_back(new IntPlanetContext(type.getCurrent()));
}


void
doPlanetAutoTaskScreen()
{
    WPlanetAutoTaskType at(global_planet_type, IntExecutionContext::pkPlanetTask);
    WPlanetAutoTaskScreen screen(at);
    at.init();
    screen.runScreen();
    at.done();
}
