/**
  *  \file client/tiles/planetecon.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/planetecon.h"
#include "game/planet.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "game/host.h"
#include "game/config.h"
#include "ui/util.h"
#include "game/cmds.h"
#include "game/global.h"
#include "util/misc.h"
#include "client/wcolorframe.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"

/*************************** WPlanetEconomyTile **************************/

/** \class WPlanetEconomyTile
    \brief Planet Economy Tile

    Displays planetary economy and resources.
    <pre>
    Mineral Mines:     XX
    Factories:         XX
    Defense Posts:     XX
    Neutronium:        XX kt
    Tritanium:         XX kt
    Duranium:          XX kt
    Molybdenum:        XX kt
    Supplies:          XX kt
    </pre> */

WPlanetEconomyTile::WPlanetEconomyTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, font_heights[FONT_NORMAL] * 8), type),
      cf_g(0)
{
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(0,  30, 20, 20), h.add(new UIButton("B", 'b')));
    addControl(GfxRect(25, 30, 20, 20), h.add(new UIButton("M", 'm')));
    addControl(GfxRect(0,  55, 20, 20), h.add(new UIButton("S", 's')));
    addControl(GfxRect(25, 55, 20, 20), h.add(new UIButton("D", 'd')));
    addControl(GfxRect(25, 90, 20, 20), h.add(new UIButton("C", 'c')));

    if (host.isPHost() && config.CPEnableGive()) {
        /* Give */
        cf_g = &h.add(new WColorFrame());
        cf_g->add(h.add(new UIButton("G", 'g')));
        addControl(GfxRect(23, 0, 24, 24), *cf_g);
    }
}

void
WPlanetEconomyTile::drawPlanetEconomyTile(UIAbstractTableWriter& w)
{
    w.newLColumn();
    w.writeFixedCell(UIColor::tc_Static, _("Mineral Mines:"));
    w.writeFixedCell(UIColor::tc_Static, _("Factories:"));
    w.writeFixedCell(UIColor::tc_Static, _("Defense Posts:"));
    w.writeFixedCell(UIColor::tc_Static, _("Neutronium:"));
    w.writeFixedCell(UIColor::tc_Static, _("Tritanium:"));
    w.writeFixedCell(UIColor::tc_Static, _("Duranium:"));
    w.writeFixedCell(UIColor::tc_Static, _("Molybdenum:"));
    w.writeFixedCell(UIColor::tc_Static, _("Supplies:"));

    GPlanet* p = dynamic_cast<GPlanet*>(getObject());
    if (p != 0 && p->hasPlanetData()) {
        w.newRColumn(6*8);
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_Mines)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_Factories)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getStructures(str_Defense)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargoRaw(el_Neutronium)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargoRaw(el_Tritanium)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargoRaw(el_Duranium)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargoRaw(el_Molybdenum)));
        w.writeVariableCell(UIColor::tc_Green, itoa(p->getCargoRaw(el_Supplies)));
    
        w.newLColumn();
        w.writeFixedCell(UIColor::tc_Green, "");
        w.writeFixedCell(UIColor::tc_Green, "");
        w.writeFixedCell(UIColor::tc_Green, "");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
        w.writeFixedCell(UIColor::tc_Green, "kt");
    }

    w.finish();
}

void
WPlanetEconomyTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 0), FONT_NORMAL);
    drawPlanetEconomyTile(tw);

    /* "give" button */
    if (cf_g != 0) {
        TFrameColor give_color = cf_None;
        if (GPlanet* p = dynamic_cast<GPlanet*>(getObject())) {
            GCommandContainer& con = getDisplayedTurn().getCommands(p->getOwner());
            if (con.getCommand(GCommand::phc_GivePlanet, p->getId()) != 0)
                give_color = cf_Yellow;
        }
        cf_g->setColor(give_color);
    }
}



WNarrowPlanetEconomyTile::WNarrowPlanetEconomyTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(250, 3*font_heights[FONT_NORMAL]), type)
{ }

void
WNarrowPlanetEconomyTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());

    GPlanet* pl = dynamic_cast<GPlanet*>(getObject());
    if (!pl) {
        drawBackground(ctx, getInnerSize(0, 0));
        return;
    }

    enum { N = 3 };
    static const TPlanetaryStructure strs[N] = { str_Mines, str_Factories, str_Defense };

    UITableWriter tw(ctx, getInnerSize(0, 0), FONT_NORMAL);
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Mines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Factories:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Defense:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999"));
    for (int i = 0; i < N; ++i) {
        if (pl->getStructures(strs[i]).isKnown()) {
            tw.writeVariableCell(UIColor::tc_Green, itoa(pl->getStructures(strs[i])));
        } else {
            tw.writeVariableCell(UIColor::tc_Green, string_t());
        }
    }

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Money:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Supplies:"));

    tw.newRColumn(fonts[FONT_NORMAL]->getTextWidth("999,999"));
    if (pl->isCargoAmountKnown(el_Money))
        tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getCargoRaw(el_Money)));
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());

    if (pl->isCargoAmountKnown(el_Supplies))
        tw.writeVariableCell(UIColor::tc_Green, numToString(pl->getCargoRaw(el_Supplies)));
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());

    tw.newLColumn();
    if (pl->isCargoAmountKnown(el_Money))
        tw.writeVariableCell(UIColor::tc_Green, "mc");
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());

    if (pl->isCargoAmountKnown(el_Money))
        tw.writeVariableCell(UIColor::tc_Green, "kt");
    else
        tw.writeVariableCell(UIColor::tc_Green, string_t());
}
