/**
  *  \file client/tiles/shipmovement.cc
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/tiles/shipmovement.h"
#include "game/mission.h"
#include "game/ship.h"
#include "game/global.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "ui/uiglobal.h"
#include "game/formula.h"
#include "game/host.h"
#include "ui/widgets/button.h"

/*************************** WShipMovementTile ***************************/

WShipMovementTile::WShipMovementTile(GObjectType& type)
    : WObjectTypeObserverTile(GfxPoint(315, 7*font_heights[FONT_NORMAL]), type)
{
    const int he = 7*font_heights[FONT_NORMAL];
    setLayout(RightJustified, 0, false);
    addControl(GfxRect(0, he - 20, 20, 20), h.add(new UIButton("A", 'a')));
    addControl(GfxRect(0, he - 45, 20, 20), h.add(new UIButton("W", 'w')));
}

void
WShipMovementTile::drawShipMovementTile(UIAbstractTableWriter& tw, bool fixed_only)
{
    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("Location:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Waypoint:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Distance:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Warp Factor:"));
    tw.writeFixedCell(UIColor::tc_Static, _("E.T.A.:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Fuel Usage:"));
    tw.writeFixedCell(UIColor::tc_Static, _("Engine Load:"));

    if (fixed_only)
        return;                 // save time by not computing movement info

    GShip* ps = dynamic_cast<GShip*>(getObject());
    if (!ps || ps->getKind() != GShip::CurrentShip)
        return;

    /* compute stuff */
    GShip& s = *ps;
    GShipTurnPredictor crystal_ball(getDisplayedTurn().getCurrentUniverse(), s.getId());
    crystal_ball.computeMovement();

    const bool is_hyper = s.isHyperdriving();

    tw.newLColumn();

    /* location */
    tw.writeVariableCell(UIColor::tc_Green, getDisplayedTurn().getCurrentUniverse().getLocationName(s.getPos(), locs_WW | locs_Verbose));

    /* waypoint */
    if (s.getMission() == GMission::msn_Intercept
        && s.getIntId() > 0
        && s.getIntId() <= NUM_SHIPS
        && getDisplayedTurn().getCurrentUniverse().getShip(s.getIntId()).isExistant())
    {
        tw.writeVariableCell(UIColor::tc_Green,
                             getDisplayedTurn().getCurrentUniverse().getShip(s.getIntId()).getName());
    } else {
        tw.writeVariableCell(UIColor::tc_Green,
                             getDisplayedTurn().getCurrentUniverse().getLocationName(s.getWaypoint(),
                                                                locs_WW | locs_Verbose | locs_Ships));
    }

    /* distance */
    double dist = distFromDX(s.getWaypointDX(), s.getWaypointDY());
    tw.writeVariableCell(!is_hyper
                         ? UIColor::tc_Green
                         : dist < 20
                         ? UIColor::tc_Red
                         : !host.isExactHyperjumpDistance2(int32(dist*dist+.1)) /* FIXME: this may lose precision, hence the +.1 */
                         ? UIColor::tc_Yellow
                         : UIColor::tc_Green,
                         format(_("%.2f ly"), dist));

    /* warp */
    if (s.getWarp() == 0 && (s.getWaypointDX() | s.getWaypointDY())) {
        tw.writeVariableCell(UIColor::tc_Red, _("not moving"));
    } else {
        // Borland sucks.
        string_t st;
        if (is_hyper)
            st = _("Hyperdrive");
        else if (s.getWarp() == 0)
            st = _("not moving");
        else
            st = format(_("Warp %d"), s.getWarp());
        tw.writeVariableCell(UIColor::tc_Green, st);
    }

    /* e.t.a. */
    {
        string_t eta;
        UISkin::Color c = UIColor::tc_Green;

        if (s.getWaypointDX() == 0 && s.getWaypointDY() == 0) {
            eta = _("at waypoint");
        } else if (is_hyper) {
            eta = _("unknown"); // FIXME?
        } else if (s.getWarp() == 0) {
            c = UIColor::tc_Red;
            eta = _("not moving");
        } else if (crystal_ball.getTurnCount() >= crystal_ball.MOVEMENT_TIME_LIMIT) {
            eta = _("too long");
        } else {
            eta = format(_("%d turn%!1{s%}"), crystal_ball.getTurnCount());
        }
        tw.writeVariableCell(c, eta);
    }

    /* fuel usage */
    {
        string_t fuel;
        UISkin::Color c = UIColor::tc_Green;

        fuel = format(_("%d kt"), crystal_ball.getMovementFuelUsed());
        if (crystal_ball.getTurnCount() >= crystal_ball.MOVEMENT_TIME_LIMIT)
            c = UIColor::tc_Yellow;
        else if (crystal_ball.getMovementFuelUsed() > s.getCargo(el_Neutronium))
            c = UIColor::tc_Red;

        // FIXME: cloak fuel
        tw.writeVariableCell(c, fuel);
    }

    /* load */
    int towee_mass = 0;
    if (s.getMission() == GMission::msn_Tow
        && s.getTowId() > 0 && s.getTowId() <= NUM_SHIPS
        && getDisplayedTurn().getCurrentUniverse().getShip(s.getTowId()).isExistant())
    {
        towee_mass = getDisplayedTurn().getCurrentUniverse().getShip(s.getTowId()).getMass();
    }
    tw.writeVariableCell(UIColor::tc_Green,
                         format(_("%d kt"), s.getMass() + towee_mass));

    tw.finish();
}

void
WShipMovementTile::drawData(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can).useColorScheme(getSkin());
    UITableWriter tw(ctx, getInnerSize(5, 20), FONT_NORMAL);
    drawShipMovementTile(tw, false);
}

