/**
  *  \file client/widgets/vcrinfomain.cc
  *  \brief WVcrInfoMain widget
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/vcrinfomain.h"
#include "client/widgets/vcrhulllist.h"
#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "ui/richdocument.h"
#include "ui/util.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "util/misc.h"
#include "game/config.h"
#include "game/formula.h"
#include "client/scr-sim.h"
#include "game/sim.h"
#include "util/recode.h"
#include "ui/msgbox.h"

static const char*const headings[] = {
    N_("Primary:"),
    N_("Secondary:"),
    N_("Tech level:"),
    N_("Mass:"),
    N_("Cargo:"),
    N_("Fuel:"),
    N_("Engines:"),
    N_("Crew:"),
};

static int
identifyEngine(const TVcr& vcr, int side, int assumedHull)
{
    // Don't guess if we don't know the hull
    if (assumedHull <= 0) {
        return 0;
    }

    // Compute effective ESB.
    // FIXME: This is the same algorithm as in PCC 1.x. Check it!
    const TVcrObject& o = vcr.objects[side];
    int32_t esb;
    if (config.AllowEngineShieldBonus()
        && (side == 0
            || vcr.battle_type == 0
            || config.AllowESBonusAgainstPlanets()))
    {
        esb = config.EngineShieldBonusRate(o.owner);
    } else {
        esb = 0;
    }

    if (config.NumExperienceLevels() > 0 && o.experience_level > 0) {
        esb += config.EModEngineShieldBonusRate(o.experience_level);
    }

    // Figure out mass that must be accounted for by ESB
    int32_t massDiff = vcr.mass[side] - getHull(assumedHull).getMass();
    if (raceId(o.owner) == 1) {
        // Scotty bonus
        massDiff -= 50;
    }

    // Is 360 kt bonus applicable?
    // FIXME: this looks wrong in PCC 1.x.
    bool is360 = (vcr.mass[side] > 140+360 && o.bay_count > 0);

    int result = 0;
    for (int i = 1; i <= NUM_ENGINES; ++i) {
        int32_t thisESB = esb * getEngine(i).getCost().get(el_Money) / 100;
        int32_t remain = massDiff - thisESB;
        if (remain == 0 || (is360 && remain == 360)) {
            if (result != 0) {
                return 0;
            } else {
                result = i;
            }
        }
    }
    return result;
}

/** Generate ship information.
    \param doc [out] Information is written here
    \param e [in] VCR
    \param side [in] Side
    \param assumedHull [in] Assume this hull type */
static void
showShipInfo(RichDocument& doc, const GVcrEntry& e, const int side, const int assumedHull)
{
    const TVcr& v = e.getVcr();
    const GHull& h = getHull(assumedHull ? assumedHull : 1);
    const int x1 = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(headings, countof(headings))) + font_ems[FONT_NORMAL];
    const int x2 = x1 + 16*font_ems[FONT_NORMAL];

    doc.addAt(x1, RichText(_("Ship")).withAttribute(RichTextStyleAttribute::Underline));
    doc.addAt(x2, RichText(_("Hull (max)")).withAttribute(RichTextStyleAttribute::Underline));
    doc.addNewline();

    // Primary weapon
    doc.add(_(headings[0]));
    if (v.objects[side].beam_count == 0) {
        doc.addAt(x1, _("none"));
    } else {
        doc.addAt(x1, string_t(format("%d " UTF_TIMES " %s", v.objects[side].beam_count, getBeam(v.objects[side].beam_type).getName())));
    }
    if (assumedHull > 0) {
        if (h.getMaxBeams() > 0) {
            doc.addAt(x2, string_t(format(_("%d beam%!1{s%}"), h.getMaxBeams())));
        } else {
            doc.addAt(x2, _("none"));
        }
    }
    doc.addNewline();

    // Secondary weapon
    doc.add(_(headings[1]));
    if (v.objects[side].bay_count > 0) {
        doc.addAt(x1, string_t(format(_("%d fighter bay%!1{s%}"), v.objects[side].bay_count)));
    } else if (v.torp_launcher_count[side] > 0) {
        doc.addAt(x1, string_t(format("%d " UTF_TIMES " %s", v.torp_launcher_count[side], getTorp(v.objects[side].torp_type).getName())));
    } else {
        doc.addAt(x1, _("none"));
    }
    if (assumedHull > 0) {
        if (h.getNumBays() > 0) {
            doc.addAt(x2, string_t(format(_("%d fighter bay%!1{s%}"), h.getNumBays())));
        } else if (h.getMaxLaunchers() > 0) {
            doc.addAt(x2, string_t(format(_("%d launcher%!1{s%}"), h.getMaxLaunchers())));
        } else {
            doc.addAt(x2, _("none"));
        }
    }
    doc.addNewline();

    // Ammo
    if (v.objects[side].bay_count > 0) {
        doc.addAt(x1, string_t(format(_("%d fighter%!1{s%}"), v.num_fighters[side])));
    } else if (v.torp_launcher_count[side] > 0) {
        doc.addAt(x1, string_t(format(_("%d torpedo%!1{es%}"), v.num_torps[side])));
    } else {
        // nix
    }
    doc.addNewline();

    // Crew
    doc.add(_(headings[7]));
    doc.addAt(x1, numToString(v.objects[side].crew));
    if (assumedHull > 0) {
        doc.addAt(x2, numToString(h.getCrew()));
    }
    doc.addNewline();

    // Tech level
    doc.add(_(headings[2]));
    if (assumedHull > 0) {
        doc.addAt(x2, numToString(h.getTech()));
    }
    doc.addNewline();

    // Mass
    doc.add(_(headings[3]));
    doc.addAt(x1, string_t(format("%d kt", v.mass[side])));
    if (assumedHull > 0) {
        doc.addAt(x2, string_t(format("%d kt", h.getMass())));
    }
    doc.addNewline();

    // Cargo
    doc.add(_(headings[4]));
    if (assumedHull > 0) {
        doc.addAt(x2, string_t(format("%d kt", h.getCargo())));
    }
    doc.addNewline();

    // Fuel
    doc.add(_(headings[5]));
    if (assumedHull > 0) {
        doc.addAt(x2, string_t(format("%d kt", h.getFuel())));
    }
    doc.addNewline();

    // Engines
    doc.add(_(headings[6]));
    if (int eng = identifyEngine(e.getVcr(), side, assumedHull)) {
        doc.addAt(x1, getEngine(eng).getName());
    } else {
        doc.addAt(x1, _("unknown"));
    }
    if (assumedHull > 0) {
        doc.addAt(x2, numToString(h.getNumEngines()));
    }
    doc.addNewline();
}

/** Add ship to simulation.
    \param ss Simulation state
    \param vcr VCR data to use
    \param side Side to use
    \param assumedHull Assumed hull, can be 0 */
static void
addShipToSim(GSimState& ss, const TVcr& vcr, int side, int assumedHull)
{
    // Add or replace ship
    const TVcrObject& o = vcr.objects[side];
    GSimShip* sh = ss.getShipById(o.id);
    if (!sh) {
        sh = &ss.getShip(ss.addShip());
    }

    // Configure ship
    sh->setId(o.id);
    sh->setName(convertGameToUtf8(getBasicString(o.name)));
    sh->setFCode("?""?""?");
    sh->setDamage(o.damage);
    sh->setShield(vcr.shield[side]);
    sh->setOwner(o.owner);
    sh->setExperienceLevel(o.experience_level);
    sh->setFlags(0);
    sh->setFlakRatingOverride(0);
    sh->setFlakCompensationOverride(0);
    sh->setCrew(o.crew);
    sh->setHull(assumedHull);
    sh->setMass(vcr.mass[side]);
    sh->setBeamType(o.beam_type);
    sh->setBeamCount(o.beam_count);
    sh->setTorpType(o.torp_type);
    sh->setTorpLauncherCount(vcr.torp_launcher_count[side]);
    sh->setBayCount(o.bay_count);
    sh->setAmmo(vcr.torp_launcher_count[side] > 0 ? vcr.num_torps[side] : vcr.num_fighters[side]);
    int en = identifyEngine(vcr, side, assumedHull);
    sh->setEngineType(en ? en : 1);
    sh->setAggressiveness(GSimShip::agg_Kill);
    sh->setInterceptId(0);
}

/******************************** Planets ********************************/

namespace {
    struct Interval {
        int min, max;

        Interval()
            : min(1), max(0)
            { }
        void include(int n);
        void set(int n);
        string_t toString(int low, int high);
    };

    struct PlanetInformation {
        bool have_base;
        int  tech_state;
        Interval pdef, sbdef, sbfighters;

        bool init(const GVcrEntry& e, int side);
    };
}

/** Include value in interval. */
void
Interval::include(int n)
{
    if (min <= max) {
        // Interval was not empty
        if (min > n)
            min = n;
        if (max < n)
            max = n;
    } else {
        // Interval was empty
        min = max = n;
    }
}

inline void
Interval::set(int n)
{
    min = max = n;
}

/** Format interval into string.
    \param low,high Maximum bounds known to user */
string_t
Interval::toString(int low, int high)
{
    // Force into range
    if (min < low)
        min = low;
    if (max > high)
        max = high;

    if (min > max) {
        // Interval is empty
        return string_t(_("impossible"));
    } else if (min == max) {
        // Single value
        return numToString(min);
    } else if (max == high && min != low) {
        // Up to maximum
        return ::format(UTF_GEQ " %d", min);
    } else if (min == low && max != high) {
        // Down to minimum
        return ::format(UTF_LEQ " %d", max);
    } else {
        // General range, or full interval
        return ::format("%d ... %d", min, max);
    }
}

bool
PlanetInformation::init(const GVcrEntry& e, const int side)
{
    const TVcrObject& o = e.getVcr().objects[side];
    int mass = e.getVcr().mass[side];
    mass -= 100;

    /* Unarmed? */
    if (mass <= 0) {
        if (o.beam_count > 0) {
            /* Huh? */
            return false;
        } else {
            /* OK */
            have_base = false;
        }
    }

    /* Derived information */
    /* Fighters  = Round(Sqrt(PD)) + BaseFighters
       BeamCount = Round(Sqrt((PD+BD)/3))
       BeamType  = Round(Sqrt(PD/2)), at least BeamTech */
    /* Validate beam count */
    int beamCount = roundToInt(std::sqrt(mass / 3.0));
    if (beamCount <= 10 && o.beam_count != beamCount) {
        return false;
    }

    /* Assume no starbase, and compute effective beam type and fighter count.
       If they do not match, there is a base. */
    int beamType = roundToInt(std::sqrt(mass / 2.0));
    if (beamType > 10) {
        beamType = 10;
    }
    int fighterCount = roundToInt(std::sqrt(double(mass)));
    int actualFighterCount = e.getVcr().num_fighters[side];

    if (o.beam_type != beamType || actualFighterCount != o.bay_count || o.bay_count != fighterCount) {
        /* Initialize */
        have_base = true;
        tech_state = 0;

        /* We're pretty certain that there is a base. Try to figure out
           the PD/SBD split. PCC 1.x tried to compute the values directly,
           but didn't catch all cases. So we just use brute force now.
           It takes about 40k cycles maximum. */
        int maxSBD = config.MaximumDefenseOnBase(o.owner);
        int maxSBFighters = config.MaximumFightersOnBase(o.owner);
        for (int sbd = 0; sbd <= mass && sbd <= maxSBD; ++sbd) {
            // Compute derived information
            int pd = mass - sbd;
            int bt = roundToInt(std::sqrt(pd / 2.0));
            if (bt > 10) {
                bt = 10;
            }
            int ftr = roundToInt(std::sqrt(double(pd)));

            // Is this possible?
            // - Must not exceed reported type/count
            // - Consider that bases sometimes add 5 bays, sometimes not
            if (bt <= o.beam_type
                && ftr <= actualFighterCount
                && (ftr == o.bay_count || ftr + 5 == o.bay_count)
                && actualFighterCount - ftr <= maxSBFighters)
            {
                if (bt < o.beam_type) {
                    tech_state |= 1;
                } else {
                    tech_state |= 2;
                }
                pdef.include(pd);
                sbdef.include(sbd);
                sbfighters.include(actualFighterCount - ftr);
            }
        }
        return true;
    } else {
        /* No hint that this might be a starbase */
        have_base = false;
        pdef.set(mass);
        tech_state = 0;
        sbdef.set(0);
        sbfighters.set(0);
        return true;
    }
}

static const char*const planetHeadings[] = {
    N_("Combat Mass:"),
    N_("Starbase:"),
    N_("Planetary Defense:"),
    N_("Starbase Defense:"),
    N_("Starbase Fighters:"),
    N_("Beam Tech:"),
};

/** Generate planet information.
    \param doc [out] Information is written here
    \param e [in] VCR
    \param side [in] Side */
static void
showPlanetInfo(RichDocument& doc, const GVcrEntry& e, const int side)
{
    const TVcrObject& o = e.getVcr().objects[side];
    int mass = e.getVcr().mass[side];

    const int x0 = font_ems[FONT_NORMAL]*2;
    const int x1 = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(planetHeadings, countof(planetHeadings))) + x0 + font_ems[FONT_NORMAL];

    /* Initial information */
    doc.add(_(planetHeadings[0]));
    doc.addAt(x1, string_t(format(_("%d kt"), mass)));
    doc.addParagraph();

    mass -= 100;
    if (mass <= 0) {
        doc.add(_("This planet is unarmed."));
        return;
    }

    doc.add(_("Derived Information:"));
    doc.addNewline();
    doc.setLeftMargin(x0);

    /* Compute derived information */
    PlanetInformation pi;
    if (!pi.init(e, side)) {
        doc.add(RichText(UIColor::tc_Red, _("Formula error.")));
        return;
    }

    doc.add(_(planetHeadings[1]));
    doc.addAt(x1, pi.have_base ? _("yes") : _("no"));
    doc.addNewline();

    doc.add(_(planetHeadings[2]));
    doc.addAt(x1, pi.pdef.toString(0, 10000));
    doc.addNewline();

    if (pi.have_base) {
        /* Base present: additional info */
        int maxSBD = config.MaximumDefenseOnBase(o.owner);
        int maxSBFighters = config.MaximumFightersOnBase(o.owner);

        doc.add(_(planetHeadings[3]));
        doc.addAt(x1, pi.sbdef.toString(0, maxSBD));
        doc.addNewline();

        doc.add(_(planetHeadings[4]));
        doc.addAt(x1, pi.sbfighters.toString(0, maxSBFighters));
        doc.addNewline();

        doc.add(_(planetHeadings[5]));
        if (pi.tech_state == 1) {
            // We know its the tech level which sets the value
            doc.addAt(x1, numToString(o.beam_type));
        } else if (pi.tech_state == 2) {
            // We know its the DPs which set the value, so tech is less
            doc.addAt(x1, string_t(format(UTF_LEQ " %d", o.beam_type)));
        } else {
            doc.addAt(x1, _("unknown"));
        }
        doc.addNewline();
    }
}

/** Add planet to simulation.
    \param ss Simulator state
    \param e VCR entry (before fight)
    \param vcr VCR data to use (before or after fight)
    \param side Side to use */
static void
addPlanetToSim(GSimState& ss, const GVcrEntry& e, const TVcr& vcr, int side)
{
    /* Use /e/ to find out the planet's initial spec. Then, use /vcr/
       to generate the simulation participant.

       FIXME: damage handling is non-existant. We would have to scale
       planet/base defense somehow, but this differs in HOST and PHost. */
    PlanetInformation pi;
    if (!pi.init(e, side)) {
        messageBox(_("I was unable to figure out this planet's equipment."),
                   _("Add to Simulation"));
        return;
    }

    const TVcrObject& o = vcr.objects[side];

    // Add planet
    ss.addPlanet();
    GSimPlanet& sp = ss.getPlanet();

    // Configure it
    sp.setId(o.id);
    sp.setName(convertGameToUtf8(getBasicString(o.name)));
    sp.setFCode("ATT");
    sp.setDamage(o.damage);
    sp.setShield(vcr.shield[side]);
    sp.setOwner(o.owner);
    sp.setExperienceLevel(o.experience_level);
    sp.setFlags(0);
    sp.setFlakRatingOverride(0);
    sp.setFlakCompensationOverride(0);

    // We use the maximum possible planet defense, and the minimum
    // possible starbase defense. The rest follows from that.
    sp.setDefense(pi.pdef.max);
    if (pi.have_base) {
        // Defense, beams
        sp.setBaseDefense(pi.sbdef.min);
        int bt = roundToInt(std::sqrt(pi.pdef.max / 2.0));
        if (bt < o.beam_type) {
            sp.setBaseBeamTech(o.beam_type);
        } else {
            sp.setBaseBeamTech(1);
        }

        // FIXME: starbase torps!
        sp.setBaseTorpTech(1);
        sp.setBaseTorps(1, 0);

        // Fighters. When using "after" data, starbase fighters
        // might have been destroyed.
        int ftr = vcr.num_fighters[side] - roundToInt(std::sqrt(double(pi.sbdef.min)));
        if (ftr < 0) {
            sp.setBaseFighters(0);
        } else {
            sp.setBaseFighters(ftr);
        }
    }
}

/****************************** WVcrInfoMain *****************************/

WVcrInfoMain::WVcrInfoMain(WVcrHullList& list)
    : list(list),
      conn_moved(observe(list.sig_moved))
{ }

WVcrInfoMain::~WVcrInfoMain()
{ }

void
WVcrInfoMain::drawContent(GfxCanvas& can)
{
    // Create blank document
    RichDocument doc;
    doc.setPageWidth(getExtent().w);

    // Generate info
    const GVcrEntry& e = list.getVcrEntry();
    int side = list.getSide();
    if (e.isPlanet(side)) {
        // Planet info
        showPlanetInfo(doc, e, side);
    } else {
        // Ship info
        showShipInfo(doc, e, side, list.getCurrentHull());
    }

    // Finish and draw
    doc.finish();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.useColorScheme(getSkin());
    drawBackground(ctx, getExtent());

    doc.draw(ctx, getExtent(), 0);
    can.update(getExtent());
}

void
WVcrInfoMain::getLayoutInfo(LayoutInfo& info)
{
    info.pref_size = info.min_size = font_sizes[FONT_NORMAL].scaledBy(35, 10);
    info.flags = 0;
}

void
WVcrInfoMain::addToSim(bool afterFight)
{
    GSimState& ss = WSimScreen::sim_state;
    if (afterFight) {
        // Requested after fight, so first run the fight
        GVcrEntry& e = list.getVcrEntry();
        GClassicVcrDatabase& db = list.getVcrDatabase();
        const int side = list.getSide();
        if (e.getStatus() == 0) {
            if (Ptr<VcrPlayer> ply = db.createPlayer()) {
                ply->setVisualizer(new VcrNullVisualizer());
                TVcr vcr_tmp = e.getVcr();
                if (ply->setCapabilities(db.getCapabilities()) && ply->checkVcr(vcr_tmp)) {
                    ply->playVcr(vcr_tmp);
                    e.setResultFromPlayer(*ply);
                }
            }
        }

        // Did we actually run the fight?
        if (e.getStatus() <= 0) {
            messageBox(_("This fight could not be played. You can use [Shift+Ins] to use this unit's "
                         "status before the fight for simulation."),
                       _("Add to Simulation"));
            return;
        }

        // Did the unit survive the fight?
        if (e.getStatus() & (side ? VcrPlayer::VCRS_RIGHT_DESTROYED : VcrPlayer::VCRS_LEFT_DESTROYED)) {
            messageBox(_("This unit did not survive the fight. You can use [Shift+Ins] to use this unit's "
                         "status before the fight for simulation."),
                       _("Add to Simulation"));
            return;
        }

        // Add it
        if (e.isPlanet(side)) {
            addPlanetToSim(ss, e, e.getResult(), side);
        } else {
            addShipToSim(ss, e.getResult(), side, list.getCurrentHull());
        }
    } else {
        // Before fight
        const GVcrEntry& e = list.getVcrEntry();
        const int side = list.getSide();
        if (e.isPlanet(side)) {
            addPlanetToSim(ss, e, e.getVcr(), side);
        } else {
            addShipToSim(ss, e.getVcr(), side, list.getCurrentHull());
        }
    }

    ss.doUpdates();
}
