/**
  *  \file game/pvcr.cc
  *  \brief PVCR Implementation
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/pvcr.h"
#include "game/config.h"
#include "game/spec.h"
#include "game/formula.h"

/** \class VcrPlayerPHost
    \brief PHost VCR Player

    This is the PHost VCR player, started as a port from the PCC
    v1.0.15 player, and went through serious tweaking.

    This player supports all PHost versions from 2.4 upwards, up to at
    least version 4.0.

    Things remaining to do:
    - some things should be done in subroutines (killFighter...)
    - cache weapon specs

    Believe me or not, I have not copied code off the PDK or PHost.
    The PDK is still (or, "again") faster... */

/*
  This seems to be pretty optimisation-resistant.

  All tests run on a vanilla Pentium 200. Each test is the mean of at
  least two runs.

  + original:                   AC: 15.74,  Non-AC: 3.87
  + inlining:                   AC: 15.27,  Non-AC: 3.72
  + PVCR_INTEGER:               AC: 15.25,  Non-AC: 3.74(!)
  + f_getDamage():              AC: 16.11(!) [not used]
  + caching config:             AC: 13.58,  Non-AC: 3.59
  + fighterMove:                AC: 11.91,  Non-AC: 3.41
  + beamFindNearestFighter:     AC:  6.70,  Non-AC: 2.04
  + template<int side>:         AC:  5.36,  Non-AC: 1.64  (currently not used)

  For comparison, PDK 4.2:      AC: 9.50,   Non-AC: 2.44
  PDK 4.3 (beamFindNearest.)    AC: 5.53,   Non-AC: 1.72

  The one reason why the PDK is faster is probably that the PDK
  contains for loops (for(side=0; side<2; ++side)), where this code
  contains two function calls for each phase. gcc can optimise that
  better. The speedup gained by templatizing seems to support that
  supposition.

  More optimisation: 100 test runs on Athlon XP 2800+ (2 GHz):
  (testvcr tests/pvcr/vcr5.dat 100)

  + start with:                 AC: 4.72,   Non-AC: 1.52
  + ShMovSpeed=0 for planets:   AC: 4.86,   Non-AC: 1.49  (not used)
  + fast forward:               AC: 4.82,   Non-AC: 1.45
  + randomRange100              AC: 4.53,   Non-AC: 1.30
  + uint32 seed                 AC: 4.33,   Non-AC: 1.27
  + fighterIntercept opt.: no obervable speedup in this test
    case, but 5% faster (4.4 -> 4.2) for carrier/carrier fight
  + PVCR_PREPARED_RNG           AC: 3.78,   Non-AC: 1.13

  Note that FF applies only in 2 of 15 battles of the test case; for
  those where it applies it yields a huge speedup, so I put it in.

  Some things which do not get anything:
  + Status& st = status[side]
  + side^1 instead of !side (although it should give better code;
    in theory, that is)
  + templatizing hit, computeShieldDamage, etc. upon
    AllowAlternativeCombat.
  + PRNG skip tables for Fast Forward: precompute how N iterations
    of randomRange(100) modify the seed for a subset of M of all
    seeds; if current seed is one of those and we have still N or
    more to go, make big jump instead of small step. With M*N=65536,
    this should get a theoretical speedup of 2 for the FF operation,
    but has no measurable overall effect. Tested with N=128. */

/** Bitmask of VCR capabilities supported by this player. */
static const unsigned SUPPORTED_CAP = vcrcap_DeathRays | vcrcap_Experience | vcrcap_Beams;

/** Movement timer. We check whether standoff distance has been
    reached every this many ticks (it could have been done a little
    simpler, but this way integrates neatly into the timer framework
    we need for regular interval checks). */
static const int DET_MOVEMENT_TIMER = 100;

/** Inactivity timer. We check whether combat made progress every this
    many ticks after standoff distance has been reached. A simple
    heuristic increases the value if necessary. */
static const int DET_INACTIVITY_TIMER = 5000;

#ifdef CONF_STATIC_VCR
TVcr VcrPlayerPHost::vcr;
VcrPlayerPHost::Status VcrPlayerPHost::status[2];
uint32 VcrPlayerPHost::seed;
bool VcrPlayerPHost::done;
int VcrPlayerPHost::one_f, VcrPlayerPHost::right_probab;
unsigned VcrPlayerPHost::capabilities;
VcrPlayerPHost::DetStatus VcrPlayerPHost::det_status[2];
bool VcrPlayerPHost::det_valid;
VcrPlayerPHost::time_type VcrPlayerPHost::det_timer;
#endif

/** Add with saturation at limits.
    \returns a+b, limited to range [min,max]. */
static inline int
addsl(int a, int b, int min, int max)
{
    long sum = long(a) + b;
    if (sum < min)
        return min;
    else if (sum > max)
        return max;
    else
        return sum;
}

// FIXME: we now have objects for each config option (ConfigStdOption,
// ConfigExpOption), so we should use these here instead of macros.

/** Get config /KEY/ for object on /SIDE/. Use this during
    initialisation. */
#define CONFIG1(KEY, SIDE) (config.KEY(vcr.objects[SIDE].owner))

/** Get config /KEY/ for object on /SIDE/. Use this during
    the battle. This uses the cached values. */
#define CONFIG2(KEY, SIDE) (status[SIDE].f.KEY)

/** Distance. Shortcut for later optimisatitons. */
#define DISTANCE() (long(status[1].r.obj_x) - status[0].r.obj_x)

/** Get experience-modified value for an option. */
#define CONFIG1EX(KEY, SIDE, LEVEL, MIN, MAX) \
      (addsl(CONFIG1(KEY, SIDE), (LEVEL) ? config.EMod##KEY(LEVEL) : 0, MIN, MAX))

/********************************** RNG **********************************/

/** Random Number Generator. Basic linear congruence. */
uint32
VcrPlayerPHost::random64k()
{
    seed = (13489 - seed * 259) & 65535;
    return seed;
}

#ifdef PVCR_PREPARED_RNG
/** Random Number Generator. Returns a random number uniformly in [0,max),
    where max is the parameter passed to PreparedRNG::operator=. */
int
VcrPlayerPHost::randomRange(const PreparedRNG& rng)
{
    int32 max = rng.limit;
    int32 i;
    do
        i = random64k();
    while (i >= max);
    return i / rng.divi;
}

/** Prepare random number generator. The PreparedRNG can later be
    passed to randomRange to return a number in [0,max). Doing it this
    way avoids one division instruction (or more) per RNG invocation. */
void
VcrPlayerPHost::PreparedRNG::operator=(int max)
{
    if (max == 0) {
        // Pathological case. These RNGs will never be called.
        divi = 65536U;
        limit = 65536U;
    } else {
        divi  = 65535U / max;
        limit = divi * max;
    }
}
#else
/** Random Number Generator. Returns a random number uniformly
    in [0, max) */
int
VcrPlayerPHost::randomRange(int max)
{
    int32 divisor = 65535U / max;
    int i;
    do
        i = random64k() / divisor;
    while (i >= max);
    return i;
}
#endif

/** Random Number Generator. Same as randomRange(100), optimized to
    avoid divisions. */
int
VcrPlayerPHost::randomRange100()
{
    int32 i;
    do
        i = random64k();
    while (i >= 65500);
    return i / 655;
}

/** Check for planet. Returns true iff /side/ is the planet. */
inline bool
VcrPlayerPHost::isPlanet(int side)
{
    return side && vcr.battle_type;
}

/************************** Standard "Hit" Stuff *************************/

/** \name Standard "Hit" Stuff */
//@{

#ifdef PVCR_INTEGER
/** Compute shield damage. Documented formula.
    Integer-only version: returns actual value (non-AC) or value scaled
    with mass+1. */
int32
VcrPlayerPHost::computeShieldDamageS(int expl, int kill, int mass, int side)
{
    if (config.AllowAlternativeCombat()) {
        int32 damage = (CONFIG2(ShieldDamageScaling, side) * int32(expl)
                        + CONFIG2(ShieldKillScaling, side) * int32(kill));
        return std::min(damage, int32(status[side].f.max_scaled));
    } else {
        int32 damage = radiv(CONFIG2(ShieldDamageScaling, side) * int32(expl)
                             + CONFIG2(ShieldKillScaling, side) * int32(kill),
                             mass + 1) + 1;
        if (damage > 10000)
            return 10000;
        else
            return damage;
    }
}

/** Compute hull damage. Documented formula.
    Integer-only version: returns actual value (non-AC) or value scaled
    with mass+1. */
int32
VcrPlayerPHost::computeHullDamageS(int expl, int kill, int mass, int side)
{
    if (config.AllowAlternativeCombat()) {
        int32 d = int32(expl) * CONFIG2(HullDamageScaling, side);
        return std::min(d, int32(status[side].f.max_scaled));
    } else {
        int32 d = radiv(computeShieldDamageS(expl, kill, mass, side) * CONFIG2(HullDamageScaling, side),
                       mass + 1) + 1;
        if (d > 10000)
            return 10000;
        else
            return d;
    }
}

/** Compute killed crew. Documented formula.
    Integer-only version: returns actual value (non-AC) or value scaled
    with mass+1. */
int32
VcrPlayerPHost::computeCrewKilledS(int kill, int mass, bool death, int side)
{
    if (config.AllowAlternativeCombat()) {
        return CONFIG2(CrewKillScaling, side) * int32(kill);
    } else {
        int32 rv = radiv(CONFIG2(CrewKillScaling, side) * int32(kill), mass + 1);
        if (death && rv == 0)
            return 1;
        else
            return rv;
    }
}
#else
/** Compute shield damage. Documented formula. Returns actual value. */
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeShieldDamage(int expl, int kill, int mass, int side)
{
    double damage = (CONFIG2(ShieldDamageScaling, side) * double(expl)
                     + CONFIG2(ShieldKillScaling, side) * double(kill))
        / double(mass + 1);
    if (damage > 10000)
        damage = 10000;
    if (! config.AllowAlternativeCombat)
        return (int) (damage + 1.5);
    else
        return damage;
}

/** Compute hull damage. Documented formula. Returns actual value. */
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeHullDamage(int expl, int kill, int mass, int side)
{
    if (config.AllowAlternativeCombat) {
        double d = double(expl) * CONFIG2(HullDamageScaling, side) / double(mass + 1);
        if (d > 10000)
            return 10000;
        else
            return d;
    } else {
        double d = computeShieldDamage(expl, kill, mass, side) * CONFIG2(HullDamageScaling, side) / double(mass + 1);
        if (d > 10000)
            return 10000;
        else
            return (int) (d + 1.5);
    }
}

/** Compute shield damage. Documented formula. Returns actual value. */
VcrPlayerPHost::damage_t
VcrPlayerPHost::computeCrewKilled(int kill, int mass, bool death, int side)
{
    double d = CONFIG2(CrewKillScaling, side) * double(kill) / double(mass + 1);
    if (config.AllowAlternativeCombat) {
        return d;
    } else {
        int i = int(d + 0.5);
        if (death && i == 0)
            return 1;
        else
            return i;
    }
}
#endif

/** Hit object.
    \param side  object being hit
    \param kill  anti-life (x-ray) power of weapon
    \param expl  explosive power of weapon
    \param is_death_ray true iff this weapon emits death rays. "hit"
    can't itself check for this, because a beam may end up here with
    expl=0 when fired while partially charged */
bool
VcrPlayerPHost::hit(int side, int kill, int expl, bool is_death_ray)
{
#ifdef PVCR_INTEGER
    int damage_rate = 100;
    bool over = false;          // FIXME: this variable use is too complicated
    Status& st = status[side];

    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;

    if (is_death_ray && (capabilities & vcrcap_DeathRays))
        goto death_ray;
    is_death_ray = false;

    /* Hit shields */
    if (st.r.shield_scaled > 0) {
        int32 damage_s = computeShieldDamageS(expl, kill, vcr.mass[side], side);
        if (st.r.shield_scaled <= damage_s) {
            // Shields are completely worn down by this hit
            damage_rate = (damage_s - st.r.shield_scaled) * 100 / damage_s;
            st.r.shield_scaled = 0;
        } else {
            // Shields still hold
            damage_rate = 0;
            st.r.shield_scaled -= damage_s;
        }
    }

    /* Shields are down -- do damage */
    if (damage_rate <= 0)
        return over;

    st.r.damage_scaled2 += computeHullDamageS(expl, kill, vcr.mass[side], side) * damage_rate;

    if (st.r.damage_scaled2 >= st.f.damage_limit_scaled) {
        over = true;
    } else {
     death_ray:
        if (!isPlanet(side)) {
            int32 killed_s = computeCrewKilledS(kill, vcr.mass[side], is_death_ray, side) * damage_rate;
            st.r.crew_scaled2 -= killed_s;
            if (st.r.crew_scaled2 < 50*st.f.scale) {
                st.r.crew_scaled2 = 0;
                over = true;
            }
        }
    }
    return over;
#else
    int damage_rate = 100;
    bool over = false;

    if (kill <= 0)
        kill = 1;
    if (expl <= 0)
        expl = 1;

    if (is_death_ray && (capabilities & vcrcap_DeathRays))
        goto death_ray;
    is_death_ray = false;

    /* Hit shields */
    if (status[side].r.shield > 0) {
        double damage = computeShieldDamage(expl, kill, vcr.mass[side], side);
        if (status[side].r.shield <= damage) {
            // Shields are completely worn down by this hit
            damage_rate = int((damage - status[side].r.shield) * 100 / damage);
            status[side].r.shield = 0;
        } else {
            // Shields still hold
            damage_rate = 0;
            status[side].r.shield -= damage;
        }
    }

    /* Shields are down -- do damage */
    if (damage_rate <= 0)
        return over;

    status[side].r.damage += computeHullDamage(expl, kill, vcr.mass[side], side) * damage_rate / 100.0;

    if (int(status[side].r.damage + 0.5) >= status[side].f.damage_limit) {
        over = true;
    } else {
     death_ray:
        if (!isPlanet(side)) {
            double killed = computeCrewKilled(kill, vcr.mass[side], is_death_ray, side) * damage_rate / 100.0;
            status[side].r.crew -= killed;
            if (status[side].r.crew < 0.5) {
                status[side].r.crew = 0;
                over = true;
            }
        }
    }
    return over;
#endif
}
//@}

/******************************** Fighters *******************************/

/** \name Fighters */
//@{

/** Compute bay recharge rate. Documented formula. Used for initialisation. */
int
VcrPlayerPHost::computeBayRechargeRate(int num, int side, int exp)
{
    int i = CONFIG1EX(BayRechargeBonus, side, exp, -500, 500) * num
        + CONFIG1EX(BayRechargeRate, side, exp, 0, 16384);
    return i > 1 ? i : 1;
}

/** Recharge Fighter Bays. */
void
VcrPlayerPHost::fighterRecharge(int side)
{
    register const int mx = vcr.objects[side].bay_count;
    register Status& st = status[side];
    for (int i = 0; i < mx; ++i)
        if (st.r.bay_status[i] < 1000)
            st.r.bay_status[i] += randomRange(st.f.bay_recharge);
}

/** Launch Fighters. */
void
VcrPlayerPHost::fighterLaunch(int side)
{
    register Status& st = status[side];

    /* can we launch a fighter? */
    if (vcr.num_fighters[side] == 0
        || st.r.launched_fighters >= CONFIG2(MaxFightersLaunched, side)
        || st.r.launch_countdown > 0)
        return;

    /* yes, we can! */
    int bay_mx = vcr.objects[side].bay_count;
    for (int bay = 0; bay < bay_mx; ++bay) {
        if (st.r.bay_status[bay] >= 1000) {
            for (int track = 0; track < CONFIG2(MaxFightersLaunched, side); ++track) {
                if (st.r.fighter_status[track] == 0) {
                    /* okay, we have a bay which is ready, and an empty track. */
                    st.r.fighter_status[track]     = 1;
                    st.r.fighter_x[track]          = st.r.obj_x;
                    st.r.fighter_shots_left[track] = CONFIG2(StrikesPerFighter, side);
                    st.r.bay_status[bay]           = 0;
                    st.r.launched_fighters++;
                    vcr.num_fighters[side]--;
                    st.r.launch_countdown = CONFIG2(BayLaunchInterval, side);
                    vis->startFighter(side, track);
                    setStatFighters(side, vcr.num_fighters[side]);
                    return;
                }
            }
        }
    }
}

/** Move Fighters. Takes back returned fighters. */
void
VcrPlayerPHost::fighterMove(int side)
{
    if (status[side].r.launched_fighters == 0)
        return;

    int dir = (side == 0) ? CONFIG2(FighterMovementSpeed, 0) : -CONFIG2(FighterMovementSpeed, 1);

    for (int track = 0; track < CONFIG2(MaxFightersLaunched, side); ++track)
        if (status[side].r.fighter_status[track] == 1) {
            status[side].r.fighter_x[track] += dir;
        } else if (status[side].r.fighter_status[track] == 2) {
            status[side].r.fighter_x[track] -= dir;
            if ((side == 0)
                ? status[side].r.fighter_x[track] < status[side].r.obj_x
                : status[side].r.fighter_x[track] > status[side].r.obj_x)
            {
                /* fighter comes back to baseship */
                status[side].r.launched_fighters--;
                vcr.num_fighters[side]++;
                vis->landFighter(side, track);
                status[side].r.fighter_status[track] = 0;
            }
        }
}

/** Do fighter intercepts. */
void
VcrPlayerPHost::fighterIntercept()
{
    static const int NEVER = -0x4000;

    /* Fighter intercept only happens if both have fighters */
    if (status[0].r.launched_fighters == 0 || status[1].r.launched_fighters == 0)
        return;

    /* Compute screen positions. The actual fighter intercept code compares fighter
       positions, with coarse granularity, because exact matches never happen. By
       ignoring the lowest 7 bits of the X coordinate, we effectively compare just
       the upper 9 bits. Fighter intercept processes all fighter pairs which is very
       expensive. We use the following optimisations:
       - precompute the fighter positions (this optimisation is also in PVCR)
       - process only occupied slots
       - process only situations where we know to have fighter pairs. Since we have
         just 9 bits that specify a position, we can easily keep track of which
         points have a left fighter, and only process right fighters at these points.
         Note that eliminating left fighters does not work even if we know that they
         have no counterpart because we must call randomRange() an appropriate number
         of times. */
    int lmatch[VCR_MAX_FTRS], rmatch[VCR_MAX_FTRS];
    int lslot[VCR_MAX_FTRS],  rslot[VCR_MAX_FTRS];
    int lcount = 0,           rcount = 0;

    uint32_t bins[16];  /* That's 512 bits */
    for (int i = 0; i < 16; ++i)
        bins[i] = 0;

    /* Compute left positions and mark occupied slots */
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, 0); ++i) {
        if (status[0].r.fighter_status[i] != 0) {
            int hash = status[0].r.fighter_x[i] >> 7;
            lslot[lcount] = i;
            lmatch[lcount] = hash;
            lcount++;
            bins[(hash >> 5) & 15] |= 1 << (hash & 31);
        }
    }

    /* Compute right positions, but eliminate impossible slots */
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, 1); ++i) {
        if (status[1].r.fighter_status[i] != 0) {
            int hash = status[1].r.fighter_x[i] >> 7;
            if (bins[(hash >> 5) & 15] & (1 << (hash & 31))) {
                rslot[rcount] = i;
                rmatch[rcount] = hash;
                rcount++;
            }
        }
    }

    /* Degenerate case: no match */
    if (!rcount) {
        for (int ls = 0; ls < lcount; ++ls)
            randomRange100();
        return;
    }

    /* Full version */
    for (int ls = 0; ls < lcount; ++ls) {
        if (randomRange100() < one_f) {
            for (int rs = 0; rs < rcount; ++rs) {
                if (rmatch[rs] == lmatch[ls]) {
                    int lf = lslot[ls];
                    int rf = rslot[rs];
                    /* two fighters at the same place, intercepting */
                    /* theoretically, here is a slight imbalance. If
                       the "right" fighter survives, it can fire again
                       while a "left" one can not. Whether this is
                       relevant in practice is unknown. */
                    if (randomRange100() < right_probab) {
                        vis->fireBeam(1, rf, lf, 1, CONFIG2(FighterBeamExplosive, 1), CONFIG2(FighterBeamKill, 1));
                        status[0].r.launched_fighters--;
                        vis->killFighter(0, lf);
                        status[0].r.fighter_status[lf] = 0;
                    } else {
                        vis->fireBeam(0, lf, rf, 1, CONFIG2(FighterBeamExplosive, 0), CONFIG2(FighterBeamKill, 0));
                        status[1].r.launched_fighters--;
                        vis->killFighter(1, rf);
                        status[1].r.fighter_status[rf] = 0;
                        rmatch[rs] = NEVER;
                    }
                    break;
                }
            }
        }
    }
}

/** Fighters attack enemy. */
bool
VcrPlayerPHost::fighterAttack(int side)
{
    /* only if we have fighters */
    if (status[side].r.launched_fighters == 0)
        return false;

    int enemy_x = status[!side].r.obj_x;
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, side); ++i) {
        if (status[side].r.fighter_status[i] == 1) {
            if (std::abs(int32(status[side].r.fighter_x[i]) - enemy_x) <= CONFIG2(FighterFiringRange, side))
            {
                status[side].r.fighter_shots_left[i]--;
                if (!status[side].r.fighter_shots_left[i])
                    status[side].r.fighter_status[i] = 2;

                bool hitres = hit(!side, CONFIG2(FighterBeamKill, side), CONFIG2(FighterBeamExplosive, side), false);
                vis->fireBeam(side, i, -1, 1,
                              CONFIG2(FighterBeamExplosive, side),
                              CONFIG2(FighterBeamKill, side));
                if (hitres)
                    return true;
            } else if ((capabilities & vcrcap_Beams) != 0) {
                /* if fighter moved past enemy, retreat */
                if (side == 0
                    ? status[side].r.fighter_x[i] > enemy_x + CONFIG2(FighterFiringRange, side)
                    : status[side].r.fighter_x[i] < enemy_x - CONFIG2(FighterFiringRange, side))
                {
                    status[side].r.fighter_shots_left[i] = 0;
                    status[side].r.fighter_status[i] = 2;
                }
            }
        }
    }
    return false;
}
//@}

/********************************* Beams *********************************/

/** \name Beams */
//@{

/** Compute beam hit odds. Documented formula. Used during initialisation. */
int
VcrPlayerPHost::computeBeamHitOdds(int beam, int side, int exp)
{
    GBeam& b = getBeam(beam);
    int i = CONFIG1EX(BeamHitBonus, side, exp, -4095, 4095) * (b.getKill() + b.getDamage()) / 100
        + CONFIG1EX(BeamHitOdds, side, exp, 0, 100);
    return i < 0 ? 0 : i;
}

/** Compute beam recharge rate. Documented formula. Used during
    initialisation. */
int
VcrPlayerPHost::computeBeamRechargeRate(int beam, int side, int exp)
{
    GBeam& b = getBeam(beam);
    int i = ((b.getKill() + b.getDamage()) * CONFIG1EX(BeamRechargeBonus, side, exp, -4095, 4095)) / 100
        + CONFIG1EX(BeamRechargeRate, side, exp, 0, 16384);
    return i < 1 ? 1 : i;
}

/** Recharge beams. */
void
VcrPlayerPHost::beamRecharge(int side)
{
    register const int mx = vcr.objects[side].beam_count;
    for (int i = 0; i < mx; ++i) {
        if (status[side].r.beam_status[i] < 1000) {
            status[side].r.beam_status[i] += randomRange(status[side].f.beam_recharge);
            vis->updateBeam(side, i);
        }
    }
}

/** Find nearest-possible fighter. Returns fighter index [0,VCR_MAX_FTRS),
    or -1 if none. */
int
VcrPlayerPHost::beamFindNearestFighter(int side)
{
    int opp = !side;
    long mindist = 100000;
    int fighter = -1;

    /* When enemy does not have a fighter, don't fire at one... */
    if (status[opp].r.launched_fighters == 0)
        return -1;

    long my_x = status[side].r.obj_x;

    /* Look for a matching fighters. When we're asked to fire on
       attacking fighters only, accept only those first. */
    for (int i = 0; i < CONFIG2(MaxFightersLaunched, opp); ++i)
        if (status[opp].r.fighter_status[i] == 1
            || (!config.FireOnAttackFighters() && status[opp].r.fighter_status[i] != 0))
        {
            long d = std::abs(my_x - status[opp].r.fighter_x[i]);
            if (d < mindist) {
                mindist = d;
                fighter = i;
            }
        }

    /* No fighter found. Check retreating fighters. Any of them does.
       Interestingly enough, PHost only checks distances for the first
       pass, not for the second one. */
    if (fighter < 0 && config.FireOnAttackFighters()) {
        for (int i = 0; i < CONFIG2(MaxFightersLaunched, opp); ++i)
            if (status[opp].r.fighter_status[i] != 0) {
                fighter = i;
                break;
            }
        return fighter;
    } else {
        /* Found a fighter on first pass. Must be in range. */
        if (mindist > CONFIG2(BeamHitFighterRange, side))
            return -1;
        else
            return fighter;
    }
}

/** Fire beams on /side/. \returns true iff battle ends. */
bool
VcrPlayerPHost::beamFire(int side)
{
    int opp = !side;
    for (int beam = 0; beam < vcr.objects[side].beam_count; ++beam) {
        const GBeam& gb = getBeam(vcr.objects[side].beam_type);
        /* Can we fire at a fighter? */
        if (status[side].r.beam_status[beam] >= CONFIG2(BeamHitFighterCharge, side)) {
            /* PVCR tests `&& distance <= 100000', but that's not needed */
            bool missing = randomRange100() >= status[side].f.beam_hit_odds;
            int fighter = beamFindNearestFighter(side);
            if (fighter >= 0) {
                /* We fire at a fighter. */
                vis->fireBeam(side, -1-beam, fighter, missing ? -1 : +1,
                              gb.getDamage(), gb.getKill());
                status[side].r.beam_status[beam] = 0;
                vis->updateBeam(side, beam);
                if (!missing) {
                    vis->killFighter(opp, fighter);
                    status[opp].r.fighter_status[fighter] = 0;
                    status[opp].r.launched_fighters--;
                }
                return false;
            }
        }

        /* Can we fire at the enemy? */
        if ((status[opp].r.launched_fighters <= 0 || (capabilities & vcrcap_Beams) != 0)
            && status[side].r.beam_status[beam] >= CONFIG2(BeamHitShipCharge, side)
            && DISTANCE() <= CONFIG2(BeamFiringRange, side))
        {
            bool missing = randomRange100() >= status[side].f.beam_hit_odds;
            int kill = gb.getKill()   * (status[side].r.beam_status[beam]/10) / 100;
            int dest = gb.getDamage() * (status[side].r.beam_status[beam]/10) / 100;
            if (raceId(vcr.objects[side].owner) == 5)
                kill *= 3;

            status[side].r.beam_status[beam] = 0;
            vis->updateBeam(side, beam);

            if (!missing) {
                bool hitr = hit(opp, kill, dest, gb.getDamage() == 0);
                vis->fireBeam(side, -1-beam, -1, 1, dest, kill);
                if (hitr)
                    return true;
            } else {
                vis->fireBeam(side, -1-beam, -1, -1, dest, kill);
            }
            return false;
        }
    }
    return false;
}
//@}

/******************************* Torpedoes *******************************/

/** \name Torpedoes */
//@{

/** Compute torpedo hit odds. Documented formula. Used in initialisation. */
int
VcrPlayerPHost::computeTorpHitOdds(int torp, int side, int exp)
{
    const GTorpLauncher& tl = getTorp(torp);
    int i = ((CONFIG1EX(TorpHitBonus, side, exp, -4095, 4095) * (tl.getKill() + tl.getDamage())) / 100
             + CONFIG1EX(TorpHitOdds, side, exp, 0, 100));
    return i < 0 ? 0 : i;
}

/** Compute torpedo recharge rate. Documented formula. Used in
    initialisation. */
int
VcrPlayerPHost::computeTubeRechargeRate(int torp, int side, int exp)
{
    const GTorpLauncher& tl = getTorp(torp);
    int i = ((CONFIG1EX(TubeRechargeBonus, side, exp, -4095, 4095) * (tl.getKill() + tl.getDamage())) / 100
             + CONFIG1EX(TubeRechargeRate, side, exp, 0, 16384));
    return i < 1 ? 1 : i;
}

/** Recharge torpedo launchers. */
void
VcrPlayerPHost::torpsRecharge(int side)
{
    register const int mx = vcr.torp_launcher_count[side];
    for (int i = 0; i < mx; ++i)
        if (status[side].r.torp_status[i] < 1000) {
            status[side].r.torp_status[i] += randomRange(status[side].f.torp_recharge);
            vis->updateTorp(side, i);
        }
}

/** Fire torpedoes. */
bool
VcrPlayerPHost::torpsFire(int side)
{
    if (DISTANCE() > CONFIG2(TorpFiringRange, side) || vcr.num_torps[side] == 0)
        return false;

    /* Scaling factor for torpedo effect. Tim scales with 2 for some reason. */
    int factor = config.AllowAlternativeCombat() ? 1 : 2;

    for (int launcher = 0; launcher < vcr.torp_launcher_count[side]; ++launcher)
        if (status[side].r.torp_status[launcher] >= 1000) {
            /* we're firing a torpedo */
            int rr = randomRange100();

            vcr.num_torps[side]--;
            status[side].r.torp_status[launcher] = 0;
            vis->updateTorp(side, launcher);
            if (rr <= status[side].f.torp_hit_odds) {
                /* we hit the enemy */
                GTorpLauncher& tl = getTorp(vcr.objects[side].torp_type);
                bool hitr = hit(!side, tl.getKill() * factor, tl.getDamage() * factor, tl.getDamage() == 0);
                setStatTorpHit(side);
                vis->fireTorp(side, rr, launcher);
                return hitr;
            } else {
                /* miss */
                vis->fireTorp(side, -1-rr, launcher);
                return false;
            }
        }
    return false;
}
//@}

/******************************** Control ********************************/

/** \name Main Stuff */
//@{

/** Move objects towards each other. */
void
VcrPlayerPHost::moveObjects()
{
    long remain = DISTANCE() - config.StandoffDistance();
    if (remain <= 0)
        return;

    /* move objects. Ensure StandoffDistance is not violated */
    long move_left = std::min (remain, long(CONFIG2(ShipMovementSpeed, 0)));
    status[0].r.obj_x += move_left;
    remain -= move_left;

    if (!isPlanet(1)) {
        long move_right = std::min (remain, long(CONFIG2(ShipMovementSpeed, 1)));
        status[1].r.obj_x -= move_right;
    }
}

/** Check whether object still has offensive capabilities. */
bool
VcrPlayerPHost::canStillFight(int side)
{
    const bool drcheck = !(capabilities & vcrcap_DeathRays) || !isPlanet(!side);
    return (vcr.objects[side].beam_count > 0 && (drcheck || getBeam(vcr.objects[side].beam_type).getDamage()))
        || (vcr.num_fighters[side] > 0 && vcr.objects[side].bay_count > 0)
        || (status[side].r.launched_fighters > 0)
        || (vcr.num_torps[side] > 0 && (drcheck || getTorp(vcr.objects[side].torp_type).getDamage()));
}

/** Initialize inactivity detection. This is called at initialisation
    of the battle and resets the detector. */
void
VcrPlayerPHost::initActivityDetector()
{
    det_valid = false;
    det_timer = DET_MOVEMENT_TIMER;
}

/** Compare inactivity detector status for one side.
    \param a     DetStatus object
    \param side  Side to compare with
    \return true iff DetStatus represents current status */
bool
VcrPlayerPHost::compareDetStatus(const DetStatus& a, int side)
{
    return a.fighters == status[side].r.launched_fighters + vcr.num_fighters[side]
        && a.torps    == vcr.num_torps[side]
#ifdef PVCR_INTEGER
        && a.shield_scaled == status[side].r.shield_scaled
        && a.damage_scaled == status[side].r.damage_scaled2
        && a.crew_scaled   == status[side].r.crew_scaled2
#else
        && a.shield  == status[side].r.shield
        && a.damage  == status[side].r.damage
        && a.crew    == status[side].r.crew
#endif
        ;
}

/** Fill in inactivity detector status for one side.
    \param a     [out] DetStatus object
    \param side  Side to fill in */
void
VcrPlayerPHost::fillInDetStatus(DetStatus& a, int side)
{
    a.fighters = status[side].r.launched_fighters + vcr.num_fighters[side];
    a.torps    = vcr.num_torps[side];
#ifdef PVCR_INTEGER
    a.shield_scaled = status[side].r.shield_scaled;
    a.damage_scaled = status[side].r.damage_scaled2;
    a.crew_scaled   = status[side].r.crew_scaled2;
#else
    a.shield  = status[side].r.shield;
    a.damage  = status[side].r.damage;
    a.crew    = status[side].r.crew;
#endif
}

/** Check whether there is combat activity. This feature was added in
    PHost 4.1a to avoid infinite loops on accidentally / maliciously
    modified combat configurations. We need not link it to particular
    host versions, though, because it just avoids infinite loops and
    does not change actual results.

    This detector works by taking a snapshot of the relevant
    parameters of a ship, and periodically checking whether these
    parameters changed. In case they don't, it assumes an infinite
    loop without any progress and ends the fight. The interval between
    checks is chosen to minimize the possibility of false positives.
    It has the known deficiency of yielding false positives when
    weapon hit and recharge rates are unattractively low (recharge
    rate 2 + hit rate 1 yields one hit on average after 100'000
    ticks).

    \return true if combat shall continue, false if it shall end */
bool
VcrPlayerPHost::checkCombatActivity()
{
    /* re-check timer expired? If not, don't check. */
    if (det_timer > time)
        return true;

    /* still moving? If yes, there's progress. */
    if (DISTANCE() > config.StandoffDistance()) {
        det_timer = time + DET_MOVEMENT_TIMER;
        return true;
    }

    /* Movement has stopped. Has there been any progress since last check? */
    if (det_valid) {
        if (compareDetStatus(det_status[0], 0) && compareDetStatus(det_status[1], 1)) {
            /* No progress */
            return false;
        }
    }

    /* Combat still runs. Compute re-check time. */
    fillInDetStatus(det_status[0], 0);
    fillInDetStatus(det_status[1], 1);
    det_valid = true;

    int32 interval = DET_INACTIVITY_TIMER;
    for (int i = 0; i < 2; ++i) {
        if (det_status[i].fighters != 0) {
            int32 L = CONFIG2(BayLaunchInterval, i) + 100;  // 100 = fuzz factor for safety
            if (CONFIG2(FighterMovementSpeed, i) > 0)
                L += 2 * config.StandoffDistance() / CONFIG2(FighterMovementSpeed, i);
            if (L > interval)
                interval = L;
        }
    }
    det_timer = time + interval;
    return true;
}

/** Play one cycle. Slightly different than in PCC 1.x due to
    more-precisely specified semantics. */
bool
VcrPlayerPHost::playCycle()
{
    if (!canStillFight(0) && !canStillFight(1))
        done = true;
    else if (!checkCombatActivity())
        done = true;

    if (done)
        return false;

    ++time;
    if (status[0].r.launch_countdown > 0)
        --status[0].r.launch_countdown;
    if (status[1].r.launch_countdown > 0)
        --status[1].r.launch_countdown;

    fighterRecharge(0);
    fighterRecharge(1);
    beamRecharge(0);
    beamRecharge(1);
    torpsRecharge(0);
    torpsRecharge(1);
    fighterLaunch(0);
    fighterLaunch(1);

    if (fighterAttack(0) || fighterAttack(1)
        || torpsFire(0) || torpsFire(1)
        || beamFire(0) || beamFire(1)) {
        done = true;
    } else {
        fighterIntercept();
        fighterMove(0);
        fighterMove(1);
        moveObjects();
    }
    return true;
}

/** Initialize player.
    \param phost3 true iff PHost 3 or later algorithm */
VcrPlayerPHost::VcrPlayerPHost(bool phost3)
    : phost3(phost3)
{ }

/** Verify one side of VCR. */
bool
VcrPlayerPHost::checkVcrSide(TVcr& vcr, int side)
{
    bool err = false;
    TVcrObject& o = vcr.objects[side];

    /* owner and types must be in range */
    if (o.owner <= 0 || o.owner > 12)
        o.owner = 12;

    if (o.beam_type < 0 || o.beam_type > NUM_BEAMS) {
        o.beam_type = o.beam_count = 0;
        err = true;
    }

    if (o.torp_type < 0 || o.torp_type > NUM_TORPS) {
        o.torp_type = vcr.torp_launcher_count[side] = 0;
        err = true;
    }

    /* validate weapon counts */
    if (o.beam_count > VCR_MAX_BEAMS) {
        o.beam_count = VCR_MAX_BEAMS;
        err = true;
    }

    if (vcr.torp_launcher_count[side] > VCR_MAX_TORPS) {
        vcr.torp_launcher_count[side] = VCR_MAX_TORPS;
        err = true;
    }

    if (o.bay_count > VCR_MAX_BAYS) {
        o.bay_count = VCR_MAX_BAYS;
        err = true;
    }

    /* ensure experience level is consistent with configuration */
    if (o.experience_level) {
        if (!(capabilities & vcrcap_Experience) || o.experience_level > config.NumExperienceLevels()) {
            o.experience_level = 0;
            err = true;
        }
    }

    return err;
}

/** Check and correct VCR. Returns true iff an error was corrected,
    false if the VCR was ok. */
bool
VcrPlayerPHost::checkVcr(TVcr& vcr)
{
    bool l = checkVcrSide(vcr, 0);
    bool r = checkVcrSide(vcr, 1);
    return l || r;
}

/** Initialize VCR Player. Creates shadow copies of configuration
    parameters and pre-charges weapons. */
void
VcrPlayerPHost::initVcr(TVcr& avcr)
{
    vcr = avcr;
    if (checkVcr(vcr)) {
        status_word = VCRS_INVALID;
        done = true;
        return;
    }

    /* Initialize Playback */
    status_word = 0;
    time = 0;
    seed = vcr.random_seed;
    done = false;

    status[0].r.obj_x = -29000;
    status[1].r.obj_x = +29000;
    for (int side = 0; side < 2; ++side) {
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat())
            status[side].f.scale = vcr.mass[side] + 1;
        else
            status[side].f.scale = 1;
        status[side].r.shield_scaled  = vcr.shield[side] * int32(status[side].f.scale);
        status[side].r.crew_scaled2   = vcr.objects[side].crew * int32(status[side].f.scale) * 100;
        status[side].r.damage_scaled2 = vcr.objects[side].damage * int32(status[side].f.scale) * 100;
        status[side].f.max_scaled = int32(10000) * status[side].f.scale;
#else
        status[side].r.shield  = vcr.shield[side];
        status[side].r.crew    = vcr.objects[side].crew;
        status[side].r.damage  = vcr.objects[side].damage;
#endif

        if (vcr.objects[side].beam_count > 0) {
            status[side].f.beam_hit_odds = computeBeamHitOdds     (vcr.objects[side].beam_type, side,
                                                                   vcr.objects[side].experience_level);
            status[side].f.beam_recharge = computeBeamRechargeRate(vcr.objects[side].beam_type, side,
                                                                   vcr.objects[side].experience_level);
        } else {
            status[side].f.beam_hit_odds = 0;
            status[side].f.beam_recharge = 0;
        }
        if (vcr.torp_launcher_count[side] > 0) {
            status[side].f.torp_hit_odds = computeTorpHitOdds(vcr.objects[side].torp_type, side,
                                                              vcr.objects[side].experience_level);
            status[side].f.torp_recharge = computeTubeRechargeRate(vcr.objects[side].torp_type, side,
                                                                   vcr.objects[side].experience_level);
        } else {
            status[side].f.torp_hit_odds = 0;
            status[side].f.torp_recharge = 0;
        }
        if (vcr.objects[side].bay_count > 0) {
            status[side].f.bay_recharge = computeBayRechargeRate(vcr.objects[side].bay_count, side,
                                                                 vcr.objects[side].experience_level);
        } else {
            status[side].f.bay_recharge = 0;
        }

        // FIXME: this is still missing, but since the original
        // line is bogus (should be p.Shield, not VCR.Shield) and it works,
        // it's probably not needed.
        //   IF NOT IsPlanet(Who) AND IsFreighter(Who) THEN VCR.Shield[Who]:=0;
        // Change line below if needed.

        int charge = (vcr.shield[side] == 100) ? 1000 : 0;
        std::fill_n(status[side].r.beam_status, int(VCR_MAX_BEAMS), charge);
        std::fill_n(status[side].r.torp_status, int(VCR_MAX_TORPS), charge);
        std::fill_n(status[side].r.bay_status,  int(VCR_MAX_BAYS),  0); // !!!
        std::fill_n(status[side].r.fighter_status, int(VCR_MAX_FTRS), 0);
        std::fill_n(status[side].r.fighter_x, int(VCR_MAX_FTRS), 0);  // not needed?
        std::fill_n(status[side].r.fighter_shots_left, int(VCR_MAX_FTRS), 0);  // not needed?
        status[side].r.launch_countdown = status[side].r.launched_fighters = 0;

#ifdef PVCR_INTEGER
        /* Original comparison was:
            if (st.r.damage_scaled2 + 50 * st.f.scale >=
                st.f.damage_limit * 100 * st.f.scale)          */
        status[side].f.damage_limit_scaled =
            raceId(vcr.objects[side].owner) == 2
            ? (150 * 2 - 1) * 50 * status[side].f.scale
            : (100 * 2 - 1) * 50 * status[side].f.scale;
#else
        status[side].f.damage_limit = (raceId(vcr.objects[side].owner) == 2 ? 150 : 100);
#endif

        CONFIG2(ShieldDamageScaling, side)  = CONFIG1(ShieldDamageScaling, side);
        CONFIG2(ShieldKillScaling, side)    = CONFIG1(ShieldKillScaling, side);
        CONFIG2(HullDamageScaling, side)    = CONFIG1(HullDamageScaling, side);
        CONFIG2(CrewKillScaling, side)      = CONFIG1(CrewKillScaling, side);
        CONFIG2(MaxFightersLaunched, side)  = CONFIG1(MaxFightersLaunched, side);
        CONFIG2(StrikesPerFighter, side)    = CONFIG1(StrikesPerFighter, side);
        CONFIG2(BayLaunchInterval, side)    = CONFIG1(BayLaunchInterval, side);
        CONFIG2(FighterMovementSpeed, side) = CONFIG1(FighterMovementSpeed, side);
        CONFIG2(FighterBeamExplosive, side) = CONFIG1(FighterBeamExplosive, side);
        CONFIG2(FighterBeamKill, side)      = CONFIG1(FighterBeamKill, side);
        CONFIG2(FighterFiringRange, side)   = CONFIG1(FighterFiringRange, side);
        CONFIG2(BeamHitFighterRange, side)  = CONFIG1(BeamHitFighterRange, side);
        CONFIG2(BeamHitFighterCharge, side) = CONFIG1(BeamHitFighterCharge, side);
        CONFIG2(BeamFiringRange, side)      = CONFIG1(BeamFiringRange, side);
        CONFIG2(BeamHitShipCharge, side)    = CONFIG1(BeamHitShipCharge, side);
        CONFIG2(TorpFiringRange, side)      = CONFIG1(TorpFiringRange, side);
        CONFIG2(ShipMovementSpeed, side)    = CONFIG1(ShipMovementSpeed, side);

        if (vcr.objects[side].experience_level) {
            /* must be -1 for array indexing. Therefore, "<"
               comparison instead of "<=" */
            const int level = vcr.objects[side].experience_level - 1;
            ASSERT(level < config.NumExperienceLevels());
            ASSERT(level < 10);
            ASSERT(capabilities & vcrcap_Experience);

#define ADD_BONUS(x, min, max) \
            CONFIG2(x, side) = addsl(CONFIG2(x, side), config.EMod##x(level+1), min, max)

            ADD_BONUS(ShieldDamageScaling,   0, 32767);
            ADD_BONUS(ShieldKillScaling,     0, 32767);
            ADD_BONUS(HullDamageScaling,     0, 32767);
            ADD_BONUS(CrewKillScaling,       0, 32767);
            ADD_BONUS(MaxFightersLaunched,   0, VCR_MAX_FTRS);
            ADD_BONUS(StrikesPerFighter,     1, 100);
            // BayLaunchInterval
            ADD_BONUS(FighterMovementSpeed,  1, 10000);
            ADD_BONUS(FighterBeamExplosive,  1, 1000);
            ADD_BONUS(FighterBeamKill,       1, 1000);
            // FighterFiringRange, BeamHitFighterRange
            ADD_BONUS(BeamHitFighterCharge,  0, 1000);
            // BeamFiringRange, BeamHitShipCharge, TorpFiringRange, ShipMovementSpeed
        }
    }

    /* pre-compute fighter intercept probabilities */
    if (phost3) {
        /* PHost 3 or 4 */
        int left_f   = (100 - CONFIG1(FighterKillOdds, 0)) * CONFIG1(FighterKillOdds, 1);
        int right_f  = (100 - CONFIG1(FighterKillOdds, 1)) * CONFIG1(FighterKillOdds, 0);
        one_f        = (left_f + right_f) / 100;
        if (one_f == 0)
            right_probab = 50;
        else
            right_probab = right_f / one_f;
        // FIXME: battle.c seems to do `right_probab = left_f / one_f'
    } else {
        /* In PHost 2, combat options were not arrayized. Hence, for
           a valid pconfig, all FighterKillOdds values are the same
           and we can pick any one */
        one_f        = config.FighterKillOdds(1);
        right_probab = 50;
    }

    initActivityDetector();
    initStat();
    vis->init();
}

/** Finish up VCR. */
void
VcrPlayerPHost::doneVcr(TVcr& avcr)
{
    if (status_word == VCRS_INVALID) {
        avcr = vcr;
        return;
    }

    ASSERT(done);

    for (int side = 0; side < 2; ++side) {
#ifdef PVCR_INTEGER
        if (config.AllowAlternativeCombat()) {
            vcr.objects[side].damage = radiv(status[side].r.damage_scaled2, status[side].f.scale * 100);
            vcr.objects[side].crew   = radiv(status[side].r.crew_scaled2,   status[side].f.scale * 100);
            vcr.shield[side]         = radiv(status[side].r.shield_scaled,  status[side].f.scale);
        } else {
            vcr.objects[side].damage = radiv(status[side].r.damage_scaled2, 100);
            vcr.objects[side].crew   = radiv(status[side].r.crew_scaled2, 100);
            vcr.shield[side]         = status[side].r.shield_scaled;
        }
#else
        vcr.objects[side].damage = int(status[side].r.damage + 0.5);
        vcr.objects[side].crew   = int(status[side].r.crew + 0.5);
        vcr.shield[side]         = int(status[side].r.shield + 0.5);
#endif
        if (vcr.objects[side].damage > 100)   // FIXME: take out?
            vcr.objects[side].damage = 100;

#ifdef PVCR_INTEGER
        if (canStillFight(side) && vcr.objects[side].damage < 100)
#else
        if (canStillFight(side) && status[side].r.damage < 99.5)
#endif
        {
            for (int i = 0; i < CONFIG2(MaxFightersLaunched, side); ++i) {
                if (status[side].r.fighter_status[i]) {
                    ++vcr.num_fighters[side];
                    --status[side].r.launched_fighters;
                    vis->landFighter(side, i);
                    status[side].r.fighter_status[i] = 0;
                }
            }
            ASSERT(status[side].r.launched_fighters == 0);
        }
        status[side].r.launched_fighters = 0;
    }

    status_word = 0;

    /* Lizards fight up to 150, but explode afterwards. */
    if (vcr.objects[0].damage >= 100)
        status_word |= VCRS_LEFT_DESTROYED;
    else if (vcr.objects[0].crew <= 0)
        status_word |= VCRS_LEFT_CAPTURED;

    if (vcr.objects[1].damage >= 100)
        status_word |= VCRS_RIGHT_DESTROYED;
    else if (vcr.battle_type == 0 && vcr.objects[1].crew <= 0)
        status_word |= VCRS_RIGHT_CAPTURED;

    if (status_word & VCRS_LEFT_DESTROYED)
        vis->killObject(0);
    if (status_word & VCRS_RIGHT_DESTROYED)
        vis->killObject(1);

    if (status_word == 0)
        /* FIXME: can we guarantee that every status not caught
           by the above is a stalemate? */
        status_word = VCRS_STALEMATE;

    avcr = vcr;
}

bool
VcrPlayerPHost::setCapabilities(uint16 cap)
{
    if (cap & ~SUPPORTED_CAP)
        return false;
    capabilities = cap;
    return true;
}

/** Fast forward. The PHost 'Fast Forward' scheme is a little more
    complicated than the one for THost because most values are
    configurable, and because the PRNG is more complicated. */
void
VcrPlayerPHost::playFastForward()
{
    if (time == 0
        && vcr.objects[0].bay_count == 0 && vcr.objects[1].bay_count == 0
        && vcr.shield[0] == 100 && vcr.shield[1] == 100)
    {
        /* no carrier in play, both fully charged. Hence, no recharge
           work to be done. First interesting thing will happen when
           we are in beam range. PHost still polls the PRNG once for
           every beam and tick, as long as beams are charged above
           BeamHitFighterCharge. */
        /* FIXME: same applies if one ship is a freighter */
        long dist_now = DISTANCE();
        int target_dist = config.StandoffDistance();
        if (target_dist < CONFIG2(BeamFiringRange, 0))
            target_dist = CONFIG2(BeamFiringRange, 0);
        if (target_dist < CONFIG2(BeamFiringRange, 1))
            target_dist = CONFIG2(BeamFiringRange, 1);
        if (vcr.torp_launcher_count[0] > 0)
            if (target_dist < CONFIG2(TorpFiringRange, 0))
                target_dist = CONFIG2(TorpFiringRange, 0);
        if (vcr.torp_launcher_count[1] > 0)
            if (target_dist < CONFIG2(TorpFiringRange, 1))
                target_dist = CONFIG2(TorpFiringRange, 1);

        int speed = CONFIG2(ShipMovementSpeed, 0);
        if (!isPlanet(1))
            speed += CONFIG2(ShipMovementSpeed, 1);

        if (target_dist < dist_now && speed > 0) {
            /* We'll move towards each other at a nonzero speed */
            int adv_time = (dist_now - target_dist) / speed;
            if (adv_time > 2) {
                /* less 2 ticks, for safety */
                adv_time -= 2;

                /* advance time */
                time += adv_time;
                status[0].r.obj_x += CONFIG2(ShipMovementSpeed, 0) * adv_time;
                if (!isPlanet(1))
                    status[1].r.obj_x -= CONFIG2(ShipMovementSpeed, 1) * adv_time;

                /* advance PRNG. Although there is a formula how to
                   advance a linear congruence *fast*, we cannot use
                   it because randomRange() advances the seed twice in
                   36 out of 65536 cases. */
                /* For the simulator, a more adventurous FF scheme
                   would be to *not* advance the PRNG, and calculate
                   the initial seed backwards when the user wants to
                   see this instance. */
                int num_pulls = 0;
                if (CONFIG2(BeamHitFighterCharge, 0) <= 1000)
                    num_pulls += vcr.objects[0].beam_count;
                if (CONFIG2(BeamHitFighterCharge, 1) <= 1000)
                    num_pulls += vcr.objects[1].beam_count;
                num_pulls *= adv_time;
                while (num_pulls-- > 0)
                    randomRange100();
            }
        }
    }
}
//@}

/** \name Accessor Interface */
//@{

int
VcrPlayerPHost::getBeamStatus(int side, int id)
{
    return status[side].r.beam_status[id] / 10;
}

int
VcrPlayerPHost::getTorpStatus(int side, int id)
{
    return status[side].r.torp_status[id] / 10;
}

int
VcrPlayerPHost::getTorpCount(int side)
{
    return vcr.num_torps[side];
}

int
VcrPlayerPHost::getFighterCount(int side)
{
    return vcr.num_fighters[side];
}

int
VcrPlayerPHost::getShields(int side)
{
    /* Round, because driver code also rounds. */
#ifdef PVCR_INTEGER
    return radiv(status[side].r.shield_scaled, status[side].f.scale);
#else
    return int(status[side].r.shield + 0.5);
#endif
}

int
VcrPlayerPHost::getDamage(int side)
{
#ifdef PVCR_INTEGER
    return radiv(status[side].r.damage_scaled2, status[side].f.scale * 100);
#else
    return int(status[side].r.damage + 0.5);
#endif
}

int
VcrPlayerPHost::getCrew(int side)
{
#ifdef PVCR_INTEGER
    return radiv(status[side].r.crew_scaled2, status[side].f.scale * 100);
#else
    return int(status[side].r.crew + 0.5);
#endif
}

int
VcrPlayerPHost::getFighterX(int side, int id)
{
    return (status[side].r.fighter_x[id]) / 128 + 320;
}

int
VcrPlayerPHost::getFighterStatus(int side, int id)
{
    return status[side].r.fighter_status[id];
}

int
VcrPlayerPHost::getObjectX(int side)
{
    return (status[side].r.obj_x) / 128 + 320;
}

long
VcrPlayerPHost::getDistance()
{
    return status[1].r.obj_x - status[0].r.obj_x;
}

/** Save status. Easy mindless way, just save everything. Actually, we
    could re-compute status[].f and one_f/right_probab from the VCR
    record, but I'm too lazy to do that now. */
Ptr<VcrStatus>
VcrPlayerPHost::getStatusToken()
{
    Ptr<StatusToken> s = new StatusToken(time);

    s->status[0]    = status[0];
    s->status[1]    = status[1];
    s->seed         = seed;
    s->status_word  = status_word;
    s->done         = done;
    s->one_f        = one_f;
    s->right_probab = right_probab;
    s->vcr          = vcr;

    return s;
}

void
VcrPlayerPHost::setStatus(Ptr<VcrStatus> s)
{
    StatusToken* t = checked_cast<StatusToken*>(s.ptr());
    status[0]    = t->status[0];
    status[1]    = t->status[1];
    seed         = t->seed;
    status_word  = t->status_word;
    done         = t->done;
    one_f        = t->one_f;
    right_probab = t->right_probab;
    vcr          = t->vcr;
    time         = t->time;
}

//@}

VcrPlayerPHost::StatusToken::StatusToken(time_type t)
    : VcrStatus(t)
{ }
