/**
  *  \file game/univ.cc
  *  \brief Universe: A Container for (Almost) Everything
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/univ.h"
#include "game/planet.h"
#include "cpluslib/misc.h"
#include "game/ship.h"
#include "game/ion.h"
#include "game/host.h"
#include "game/formula.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/mine.h"
#include "util/misc.h"
#include "game/parser.h"
#include "game/stat.h"
#include "game/mission.h"


/** Update all objects in a particular object type.
    \param type The type to update */
static bool
updateType(GObjectType& type)
{
    bool changed = false;
    for (int i = type.getNextNoWrap(0, false); i != 0; i = type.getNextNoWrap(i, false)) {
        GObject& obj = type.getObject(i);
        if (obj.isDirty()) {
            obj.notifyListeners();
            changed = true;
        }
    }
    return changed;
}



/** Create blank universe. */
GUniverse::GUniverse()
    : universe_changed(false), turn_nr(0)
{
    std::fill_n(planets, countof(planets), (GPlanet*) 0);
    std::fill_n(ships,   countof(ships),   (GShip*) 0);
    std::fill_n(storms,  countof(storms),  (GIonStorm*) 0);
}

/** Destroy universe. */
GUniverse::~GUniverse()
{
    done();
}

/** Initialize universe. This fills in all objects. */
void
GUniverse::init()
{
    // FIXME: make sure that universe is actually empty
    for (int i = 0; i < NUM_PLANETS; ++i)
        planets[i] = new GPlanet(i+1);
    for (int i = 0; i < NUM_SHIPS; ++i)
        ships[i] = new GShip(i+1);
    for (int i = 0; i < NUM_STORMS; ++i)
        storms[i] = new GIonStorm(i+1);
    minefields.clear();
}

/** Uninitialize universe. */
void
GUniverse::done()
{
    for (int i = 0; i < NUM_PLANETS; ++i) {
        delete planets[i];
        planets[i] = 0;
    }
    for (int i = 0; i < NUM_SHIPS; ++i) {
        delete ships[i];
        ships[i] = 0;
    }
    for (int i = 0; i < NUM_STORMS; ++i) {
        delete storms[i];
        storms[i] = 0;
    }
    minefields.clear();
}

/** Postprocess universe. Call this to make sure structural changes
    propagate.
    \param playing_set Set of players we play. */
void
GUniverse::postprocess(const GPlayerSet playing_set, GPlayerSet data_set)
{
    this->playing_set = playing_set;
    this->data_set = data_set;
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = getPlanet(i);
        p.internalCheck();
        p.setIsPlayed(p.isExistant() && p.hasPlanetData() && playing_set.contains(p.getOwner()));
    }
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getShip(i);
        s.internalCheck();
        s.setIsPlayed(s.isExistant() && s.hasShipData() && playing_set.contains(s.getOwner()));
    }
    for (GMinefieldContainer::iterator i = minefields.begin(); i != minefields.end(); /* empty */) {
        i->internalCheck(*this);
        if (i->isExistant()) {
            ++i;
        } else {
            i = minefields.erase(i);
        }
    }
    getUfos().postprocess(*this);
    for (int i = 1; i <= NUM_SHIPS; ++i)
        getShip(i).combinedCheck1(*this);
    for (int i = 1; i <= NUM_PLANETS; ++i)
        getPlanet(i).combinedCheck2(*this);

    getDrawings().eraseExpiredDrawings(getTurnNumber());

    // FIXME: synthesize scores if a score blanker is used
}

/** Recompute planet orbit flags. */
void
GUniverse::recomputeOrbitFlags()
{
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        getPlanet(i).clearOrbitFlags();
    }

    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& sh = getShip(i);
        if (sh.isExistant() && sh.getOrbit() != 0)
            getPlanet(sh.getOrbit()).setShipInOrbit(sh);
    }
}

/** Get planet at location, with warp wells. If there is no planet at
    the location, look whether the point is in the warp well of one.
    \param pt Location, need not be normalized
    \param gravity true to look into warp well. If false, function
                   behaves exactly like getPlanetAt(int).
    \return Id of planet, or zero if none */
int
GUniverse::getPlanetAt(const GPoint pt, bool gravity) const
{
    int rv = getPlanetAt(pt);
    if (!rv && gravity)
        rv = getGravityPlanetAt(pt);
    return rv;
}

/** Get planet at location.
    \param pt Location, need not be normalized
    \return Id of planet, or zero if none */
int
GUniverse::getPlanetAt(const GPoint pt) const
{
    const GPoint real_pos = pt.getNormalized();
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (getPlanet(i).isExistant() && getPlanet(i).getPos() == real_pos)
            return i;
    return 0;
}

/** Get planet from warp well location.
    \param pt Location
    \pre getPlanetAt(pt) == 0
    \return Id of planet if pt is in its warp wells, 0 otherwise */
int
GUniverse::getGravityPlanetAt(GPoint pt) const
{
    /* easy case */
    if (!config.AllowGravityWells())
        return 0;

    if (host.isPHost()) {
        /* PHost gravity wells */
        int sqs = square(config.GravityWellRange());
        for (int i = NUM_PLANETS; i > 0; --i) {
            const GPlanet& p = getPlanet(i);
            if (p.isExistant()) {
                if (config.RoundGravityWells()) {
                    if (distanceSquared(p.getPos(), pt) <= sqs)
                        return i;
                } else {
                    GPoint p2 = p.getPos().getNearestTo(pt);
                    if (square(p2.x - pt.x) <= sqs && square(p2.y - pt.y) <= sqs)
                        return i;
                }
            }
        }
        return 0;
    } else {
        /* THost gravity wells: round, 3 ly, not wrapped, "cumulative" */
        int pid = 0;
        for (int i = 1; i <= NUM_PLANETS; ++i) {
            const GPlanet& p = getPlanet(i);
            if (p.isExistant() && distanceSquaredRaw(p.getPos(), pt) <= 9) {
                pt = p.getPos();  // (!)
                pid = i;
            }
        }
        return pid;
    }
}

/** Format name of a planet. */
static string_t
formatPlanetName(const GUniverse& univ, int pid, int flags)
{
    return format(((flags & locs_Orbit)
                   ? ((flags & locs_Verbose)
                      ? _("Orbit of %s (Planet #%d)")
                      : _("Orbit of %s (#%d)"))
                   : ((flags & locs_Verbose)
                      ? _("%s (Planet #%d)")
                      : _("%s (#%d)"))),
                  univ.getPlanet(pid).getName(), pid);
}

/** Get name of a location in human-readable form.
    \param pt     location
    \param flags  details of requested string
    - locs_Ships: also use ship names as location names, not just planets
    - locs_WW: return "near PLANETNAME" if applicable
    - locs_Orbit: say "Orbit of PLANET" if applicable
    - locs_Verbose: be more verbose */
string_t
GUniverse::getLocationName(GPoint pt, int flags) const
{
    /* is it a bird? */
    int pid = getPlanetAt(pt);
    if (pid)
        return formatPlanetName(*this, pid, flags);

    /* is it a plane? */
    if (flags & locs_Ships) {
        int sid = getAnyShipAt(pt);
        if (sid)
            return getShip(sid).getFullName();
    }
    if (flags & locs_WW) {
        pid = getGravityPlanetAt(pt);
        if (pid)
            return format(_("near %s"), formatPlanetName(*this, pid, flags & ~locs_Orbit));
    }
    if (flags & locs_NoSpace) {
        return string_t();
    }
    return format((flags & locs_Verbose) ? _("Deep Space (%d,%d)") : "(%d,%d)",
                  pt.x, pt.y);
}

/** Check whether a ship is being towed.
    \param sid Ship Id */
bool
GUniverse::isShipTowed(const int sid) const
{
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& sh = getShip(i);
        if (sid != i && sh.isPlayed() && sh.getMission() == GMission::msn_Tow && sh.getTowId() == sid) {
            return true;
        }
    }
    return false;
}

/** Get ship at position. Any race does. This is mainly used for naming
    locations, and for things like to tell whether a "L" command would succeed.
    \param pt position to check
    \returns Id number of a ship at position \c pt, or zero if none. */
int
GUniverse::getAnyShipAt(GPoint pt) const
{
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (getShip(i).isExistant() && getShip(i).getPos() == pt)
            return i;
    return 0;
}

/** Count own ships at position.
    \param pt position to check
    \returns number of currently-played ships which are at the
    specified position \c pt. */
int
GUniverse::getNumberOfPlayedShipsAt(GPoint pt) const
{
    int n = 0;
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (getShip(i).isExistant() && getShip(i).getPos() == pt && getShip(i).isPlayed())
            ++n;
    return n;
}

/** Perform all updates. This will poll all updatable objects, and
    raise the appropriate signals. */
void
GUniverse::doScreenUpdates()
{
    bool changed = false;

    /* Tell everyone we're going to do updates */
    sig_pre_update.raise();

    /* Update individual objects */
    {
        GHistoryShipType type;
        type.setUniverse(this);
        changed |= updateType(type);
    }
    {
        GAnyPlanetType type;
        type.setUniverse(this);
        changed |= updateType(type);
    }
    {
        GIonStormType type;
        type.setUniverse(this);
        changed |= updateType(type);
    }
    {
        GMinefieldType type;
        type.setUniverse(this);
        changed |= updateType(type);
    }
    {
        GUfoType type;
        type.setUniverse(this);
        changed |= updateType(type);
    }

    /* Tell everyone we did updates */
    if (changed || universe_changed)
        sig_universe_changed.raise();
    universe_changed = false;
}

GPlanet&
GUniverse::getPlanet(int id)
{
    ASSERT(id > 0 && id <= NUM_PLANETS);
    return *planets[id-1];
}

const GPlanet&
GUniverse::getPlanet(int id) const
{
    ASSERT(id > 0 && id <= NUM_PLANETS);
    return *planets[id-1];
}

bool
GUniverse::isValidPlanetId(int id) const
{
    return id > 0 && id <= NUM_PLANETS;
}

GShip&
GUniverse::getShip(int id)
{
    ASSERT(id > 0 && id <= NUM_SHIPS);
    return *ships[id-1];
}

const GShip&
GUniverse::getShip(int id) const
{
    ASSERT(id > 0 && id <= NUM_SHIPS);
    return *ships[id-1];
}

bool
GUniverse::isValidShipId(int id) const
{
    return id > 0 && id <= NUM_SHIPS;
}

GIonStorm&
GUniverse::getIonStorm(int sid)
{
    ASSERT(sid > 0 && sid <= NUM_STORMS);
    return *storms[sid-1];
}

const GIonStorm&
GUniverse::getIonStorm(int sid) const
{
    ASSERT(sid > 0 && sid <= NUM_STORMS);
    return *storms[sid-1];
}

bool
GUniverse::isValidIonStormId(int id) const
{
    return id > 0 && id <= NUM_STORMS;
}

void
GUniverse::addMessageInformation(const GMessageInformation& msgi)
{
    switch (msgi.getObjectType()) {
     case MessageShip:
        // FIXME:
        break;
     case MessagePlanet:
     case MessageStarbase:
        if (msgi.getObjectId() > 0 && msgi.getObjectId() <= NUM_PLANETS)
            getPlanet(msgi.getObjectId()).addMessageInformation(msgi);
        break;
     case MessageMinefield:
        {
            int id = msgi.getObjectId();
            int x = -1, y = -1, owner = -1, reason = GMinefield::mfa_Scanned, type = -1;
            int32_t units = -1;
            for (GMessageInformation::iterator i = msgi.begin(); i != msgi.end(); ++i) {
                if (GMessageIntValue* iv = dynamic_cast<GMessageIntValue*>(*i)) {
                    switch (iv->getIndex()) {
                     case mi_X:
                        x = iv->getValue();
                        break;
                     case mi_Y:
                        y = iv->getValue();
                        break;
                     case mi_Owner:
                        owner = iv->getValue();
                        break;
                     case mi_MineUnits:
                        units = iv->getValue();
                        break;
                     case mi_MineType:
                        type = iv->getValue();
                        break;
                     case mi_MineScanReason:
                        reason = iv->getValue();
                        break;
                     default:
                        break;
                    }
                }
            }

            /* Assimilate minefield when we have all necessary data */
            if (x > 0 && y > 0 && owner > 0 && type >= 0 && units >= 0) {
                getMinefields().getCreate(id, reason).setData(x, y, owner, units, getTurnNumber(), type != 0);
            }
        }
        break;
     case MessageNoObject:
        break;

     case MessagePlayerScore:
        /* Score. Right now, we can only process current scores, because
           database access needs a timestamp. */
        if (isStatisticsFileLoaded() && msgi.getTurnNumber() == getTurnNumber())
            getStatistics().addMessageInformation(msgi, getTimestamp());
        break;

     case MessageConfig:
        for (GMessageInformation::iterator i = msgi.begin(); i != msgi.end(); ++i) {
            if (GMessageConfigValue* cv = dynamic_cast<GMessageConfigValue*>(*i)) {
                if (ConfigOption* opt = config.getConfigOption(cv->getIndex())) {
                    try {
                        opt->set(cv->getValue());
                    }
                    catch (...) {
                        /* Ignore any error. */
                    }
                }
            }
        }
        break;
    }
}
