/**
  *  \file game/vcrdb.cc
  *  \brief VCR Database Implementation
  *
  *  As of early 2005, we support not only the HOST / PVCR
  *  distinction, but also the FLAK multi-ship combat engine which has
  *  a completely different VCR format. This module provides the
  *  database base classes, as well as the storage engine for
  *  "classic" (one-on-one) VCRs.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/vcrdb.h"
#include "util/except.h"
#include "game/tvcr.h"
#include "game/spec.h"
#include "game/pvcr.h"
#include "util/nls.h"
#include "util/misc.h"
#include "io/io.h"

/*! \class GVcrEntry

    A GVcrEntry is to a TVcr what GBeam is to TBeam: the high-level
    representation.

    For VCRs, we have a three-stage representation:
    - TRawVcr is a VCR as it appears in VCR.DAT files;
    - Since the VCR representation is ambiguous, mostly for
      PlanetsHaveTubes things, it is unpacked into a TVcr.
      Much code uses these;
    - GVcrEntry is the high-level representation for "one battle".

    Right now, there is no fine-grain accessor interface. */
GVcrEntry::GVcrEntry(const TVcr& vcr)
    : status(0),
      vcr(vcr)
{
    hulls[0] = guessHull(0);
    hulls[1] = guessHull(1);
}

GVcrEntry::~GVcrEntry()
{ }

/** Guess hull for specified side.
    \returns hull if unique, 0 if none possible or ambiguous. */
int
GVcrEntry::guessHull(int side) const
{
    /* planets don't have hulls. */
    if (isPlanet(side))
        return 0;

    /* see if PHost sent us the hull type. */
    if (vcr.objects[side].hull_or_zero)
        return vcr.objects[side].hull_or_zero;

    /* otherwise, try all hulls. */
    int type = 0;
    for (int h = 1; h <= getNumHulls(); ++h) {
        if (canBeHull(side, h)) {
            if (type == 0)
                type = h;
            else
                return 0;       // ambiguous, can't guess
        }
    }
    return type;
}

/** Check if hull matches.
    \param side     0 or 1 (left or right side of VCR)
    \param hull_id  hull id to check [1, getNumHulls()].
    \pre we're actually talking about a ship, i.e. !isPlanet(side). */
bool
GVcrEntry::canBeHull(int side, int hull_id) const
{
    const GHull& hull = getHull(hull_id);
    const TVcrObject& vcro = vcr.objects[side];

    if (vcro.hull_or_zero) {
        /* hull known - case closed */
        return (vcro.hull_or_zero == hull_id);
    } else {
        // FIXME: handle THost's easter egg here? Nebulas with Transwarps are reported
        // with wrong picture
        // FIXME: check other hull properties?
        // - mass: VCR ship shouldn't be lighter than original - true?
        // - crew: can be smaller or larger

        /* picture must match */
        if (vcro.picture != hull.getPicture())
            return false;
        /* must not have more beams/torps than hull allows */
        if (vcro.beam_count > hull.getMaxBeams()
            || vcr.torp_launcher_count[side] > hull.getMaxLaunchers())
            return false;
        /* for fighter bays, the only k.o. criterion is that ship has
           fighters but hull has not. The number of bays can be
           smaller (NTP, damage) or larger (scotty bonus). */
        if (vcro.bay_count && !hull.getNumBays())
            return false;
        return true;
    }
}

/** Get picture to use for a side. If we know the hull type, use the
    hullspec image, which might be different from the one Host gave
    us. Defined only for ships; will return 0 for planets. */
int
GVcrEntry::getShipPicture(int side) const
{
    if (isPlanet(side)) {
        return 0;
    } else if (hulls[side] != 0) {
        return ::getHull(hulls[side]).getPicture();
    } else {
        return vcr.objects[side].picture;
    }
}

/** Store result. Fetches the result from the specified
    VCR Player. The player must have called playVcr(), but
    not yet doneVcr(). */
void
GVcrEntry::setResultFromPlayer(VcrPlayer& player)
{
    player.doneVcr(result);
    status = player.getStatusWord();
}

/** Set Result. Sets the result status from the specified
    parameters.
    \param status Status word
    \param result Result VCR status */
void
GVcrEntry::setResult(int status, const TVcr& result)
{
    this->status = status;
    this->result = result;
}


/************************** GClassicVcrDatabase **************************/

/** Check whether VCR record bears the PHost magic number. */
static inline bool
hasPHostMagic(const TRawVcr& vcr)
{
    /* some docs claim that there is another magic, 65261, but I have
       never seen that one in the wild. Actually, all docs from PHost
       1.1 up to 2.13 as well as 3.x and 4.x use 48879. */
    return ((vcr.random_seed + vcr.signature) & 0xFFFFU) == 48879;
}

/** Construct blank VCR database of given type. */
GClassicVcrDatabase::GClassicVcrDatabase(Type type)
    : content(), type(type), capabilities(0)
{
    /* FIXME: this forces all PHost 4 sims to use PHost 4.0k rules. It
       remains to debate whether it makes sense to offer choice here. 
       OTOH, the change in PHost was done with the assumption that no
       sane game uses the configs where 4.0k and pre-4.0k differ. */
    if (type == vcr_PHost4)
        capabilities = vcrcap_Beams | vcrcap_DeathRays | vcrcap_Experience;
}

/** Construct VCR database from file. */
GClassicVcrDatabase::GClassicVcrDatabase(Stream& file)
    : content(), type(vcr_Host), capabilities(0)
{
    char    buffer[TRawVcr::size];
    bool    maybe_phost;
    uint16  count;

    ASSERT(&file);

    file.readT(buffer, 2);

    count = getUint16(buffer);
    maybe_phost = false;

    this->type = vcr_Host;
    this->capabilities = 0;

    while (count--) {
        TRawVcr raw_vcr;

        file.readT(buffer, sizeof(buffer));
        getStructure(buffer, raw_vcr);
        if (content.empty()) {
            /* this is the first VCR */
            maybe_phost = hasPHostMagic(raw_vcr);
        } else if (maybe_phost && !count) {
            /* this is the last VCR. Some things to note:
               - it may be PHost 2's config battle. Discard it but
                 memorize type.
               - it may be the bogus battle from CORR (all but name
                 zeroed) */
            if (hasPHostMagic(raw_vcr) && raw_vcr.battle_type >= 2) {
                /* FIXME: right now, we assume that the pconfig is
                   right. For a standalone PVCR, we should parse the
                   config structure. */
                if ((raw_vcr._planet_temp_code & 255) == 2)
                    type = vcr_PHost2;
                else
                    type = vcr_UnknownPHost;
                break;
            }
            if (raw_vcr.objects[0].owner == 0 || raw_vcr.objects[1].owner == 0)
                /* CORR */
                break;
        }

        content.push_back(new GVcrEntry(TVcr(raw_vcr)));
    }

    if (maybe_phost && type == vcr_Host) {
        type = vcr_PHost3;

        uint16 cw = content.front()->getVcr()._planet_temp_code;
        if (cw & vcrcap_Valid)
            capabilities = cw & ~vcrcap_Valid;
        if (capabilities)
            type = vcr_PHost4;
    }

    /* If it still looks like Host, it might be NuHost VCRs unpacked
       by c2nu. */
    if (type == vcr_Host && !content.empty() && content.front()->getVcr().signature == 0x554E) {
        type = vcr_NuHost;
    }
}

GClassicVcrDatabase::~GClassicVcrDatabase()
{ }

int
GClassicVcrDatabase::getCount() const
{
    return content.size();
}

void
GClassicVcrDatabase::visit(GVcrDatabaseVisitor& vis)
{
    vis.visitClassicDatabase(*this);
}

/** Get entry from file, 0-based. */
GVcrEntry&
GClassicVcrDatabase::getEntry(int n)
{
    ASSERT(n >= 0 && n < int(content.size()));
    return *content[n];
}

const GVcrEntry&
GClassicVcrDatabase::getEntry(int n) const
{
    ASSERT(n >= 0 && n < int(content.size()));
    return *content[n];
}

/** Append new entry to VCR database.
    \return the entry */
GVcrEntry&
GClassicVcrDatabase::addEntry(const TVcr& vcr)
{
    GVcrEntry* e = new GVcrEntry(vcr);
    content.push_back(e);
    return *e;
}

/** Create a VCR player that can play our VCRs.
    \return player object, or 0 if we cannot play it */
Ptr<VcrPlayer>
GClassicVcrDatabase::createPlayer()
{
    switch (type) {
     case vcr_Host:
        return new VcrPlayerTHost(false);
     case vcr_NuHost:
        return new VcrPlayerTHost(true);
     case vcr_PHost4:
        {
            Ptr<VcrPlayer> p = new VcrPlayerPHost(true);
            p->setCapabilities(capabilities);
            return p;
        }
     case vcr_PHost3:
        return new VcrPlayerPHost(true);
     case vcr_PHost2:
        return new VcrPlayerPHost(false);
     default:
        throw std::runtime_error("Invalid VCR type");
    }
}

/** Get name corresponding to a database type. */
const char*
GClassicVcrDatabase::getTypeName(Type t)
{
    static const char*const names[] = {
        N_("Host"),
        N_("Unknown"),
        N_("Unknown PHost"),
        N_("PHost 2"),
        N_("PHost 3"),
        N_("PHost 4"),
        N_("NuHost"),
    };
    return _(names[t]);
}
