/**
  *  \file gfx/bmfont.h
  *  \brief Bitmap Fonts
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GFX_BMFONT_H
#define PCC_V2_GFX_BMFONT_H

#include <vector>
#include "cpluslib/types.h"
#include "gfx/font.h"
#include "cpluslib/ptrvector.h"

class Stream;

/** Glyph for a bitmap font. Contains the bitmap of the character,
    plus the list of anti-aliasing "hints". */
class GfxBitmapGlyph {
    uint16_t width;                      /**< Width of character, in pixels. */
    uint16_t height;                     /**< Height of character, in pixels. */
    std::vector<uint8_t> data;           /**< Character data, in format for blitPattern(). */
    std::vector<uint16_t> aadata;        /**< Anti-aliasing "hints". Two elements per item, an X and Y element. */

    int getBytesPerLine() const;

 public:
    GfxBitmapGlyph();
    GfxBitmapGlyph(uint16_t width, uint16_t height);
    GfxBitmapGlyph(uint16_t width, uint16_t height, const uint8_t* data);
    ~GfxBitmapGlyph();

    void addAAHint(uint16_t x, uint16_t y);
    int getHeight() const;
    int getWidth() const;
    void draw(GfxContext& ctx, GfxPoint pt) const;
    void drawColored(GfxCanvas& can, GfxPoint pt, uint32_t pixel_color, uint32_t aa_color) const;

    void set(int x, int y, bool value);
    bool get(int x, int y) const;

    const std::vector<uint16_t>& getAAData() const;
    const std::vector<uint8_t>& getData() const;

    static uint32_t getBytesForSize(uint16_t width, uint16_t height);
};

/** Bitmap font. Contains a list of bitmaps. Can render Unicode text. */
class GfxBitmapFont : public GfxFont {
    /** Nested array of glyphs. The inner vectors contain up to 256
        character glyphs corresponding to the lower 8 bits of an
        Unicode codepoint. The outer vector is indexed by the upper 8
        bits of a codepoint. */
    ptr_vector<ptr_vector<GfxBitmapGlyph> > glyphs;

    /** Height of this font. */
    int height;
 public:
    GfxBitmapFont();
    ~GfxBitmapFont();

    // GfxBitmapFont methods:
    void addNewGlyph(uint16_t id, GfxBitmapGlyph* g);
    const GfxBitmapGlyph* getGlyph(uint16_t id) const;
    uint32_t getCurrentCharacterLimit() const;
    int getHeight() const;

    // GfxFont virtuals:
    void outText(GfxContext& ctx, GfxPoint pt, string_t text);
    int getTextWidth(string_t text);
    int getTextHeight(string_t text);
    bool setScale(int mul, int div);
    GfxPoint getScale();
    unsigned getCapabilities();

    void load(Stream& s, int index);
};


/** Get height of this glyph in pixels. */
inline int
GfxBitmapGlyph::getHeight() const
{
    return height;
}

/** Get width of this glyph in pixels. */
inline int
GfxBitmapGlyph::getWidth() const
{
    return width;
}

/** Get anti-aliasing data. */
inline const std::vector<uint16_t>&
GfxBitmapGlyph::getAAData() const
{
    return aadata;
}

/** Get pixel data. */
inline const std::vector<uint8_t>&
GfxBitmapGlyph::getData() const
{
    return data;
}

/** Get font height. */
inline int
GfxBitmapFont::getHeight() const
{
    return height;
}

#endif
