/**
  *  \file int/if/vcrif.cc
  *  \brief Interpreter Game Interface: VCRs
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/format.h"
#include "cpluslib/misc.h"
#include "game/global.h"
#include "game/racename.h"
#include "game/spec.h"
#include "game/vcrdb.h"
#include "int/args.h"
#include "int/if/ifutil.h"
#include "int/if/vcrif.h"
#include "int/interror.h"
#include "int/value.h"
#include "io/io.h"
#include "util/nls.h"
#include "util/recode.h"

IntValue*
IFVcrGet(IntArgBlock& args)
{
    int32_t id;
    args.checkArgumentCount(1);
    if (!haveDisplayedTurn())
        return 0;
    if (!checkIntArg(id, args.getNext(), 1, getDisplayedTurn().getVcrs().getCount()))
        return 0;
    return new IntVcrContext(id);
}

IntContext*
IFVcrMake()
{
    if (!haveDisplayedTurn())
        return 0;

    GVcrDatabase& db = getDisplayedTurn().getVcrs();
    if (db.getCount() == 0)
        return 0;

    return new IntVcrContext(1);
}

/***************************** IntVcrContext *****************************/

namespace { enum { VcrProperty, LeftProperty, RightProperty }; }

static const TIntMapping vcr_mapping[] = {
    { "ALGORITHM",                ivpAlgorithm,     VcrProperty },
    { "CAPABILITIES",             ivpFlags,         VcrProperty },
    { "LEFT",                     ivsNameFull,      LeftProperty },
    { "LEFT.AUX",                 ivsAuxName,       LeftProperty },
    { "LEFT.AUX$",                ivsAuxId,         LeftProperty },
    { "LEFT.AUX.AMMO",            ivsAuxAmmo,       LeftProperty },
    { "LEFT.AUX.COUNT",           ivsAuxCount,      LeftProperty },
    { "LEFT.AUX.SHORT",           ivsAuxShort,      LeftProperty },
    { "LEFT.BEAM",                ivsBeamName,      LeftProperty },
    { "LEFT.BEAM$",               ivsBeamId,        LeftProperty },
    { "LEFT.BEAM.COUNT",          ivsBeamCount,     LeftProperty },
    { "LEFT.BEAM.SHORT",          ivsBeamShort,     LeftProperty },
    { "LEFT.CREW",                ivsCrew,          LeftProperty },
    { "LEFT.CREW$",               ivsCrewRaw,       LeftProperty },
    { "LEFT.DAMAGE",              ivsDamage,        LeftProperty },
    { "LEFT.FIGHTER.BAYS",        ivsFighterBays,   LeftProperty },
    { "LEFT.FIGHTER.COUNT",       ivsFighterCount,  LeftProperty },
    { "LEFT.HULL",                ivsHullName,      LeftProperty },
    { "LEFT.HULL$",               ivsHullId,        LeftProperty },
    { "LEFT.ID",                  ivsId,            LeftProperty },
    { "LEFT.IMAGE",               ivsImage,         LeftProperty },
    { "LEFT.LEVEL",               ivsLevel,         LeftProperty },
    { "LEFT.MASS",                ivsMass,          LeftProperty },
    { "LEFT.NAME",                ivsName,          LeftProperty },
    { "LEFT.OWNER",               ivsOwnerShort,    LeftProperty },
    { "LEFT.OWNER$",              ivsOwnerId,       LeftProperty },
    { "LEFT.OWNER.ADJ",           ivsOwnerAdj,      LeftProperty },
    { "LEFT.SHIELD",              ivsShield,        LeftProperty },
    { "LEFT.STATUS",              ivsStatus,        LeftProperty },
    { "LEFT.TORP",                ivsTorpName,      LeftProperty },
    { "LEFT.TORP$",               ivsTorpId,        LeftProperty },
    { "LEFT.TORP.COUNT",          ivsTorpCount,     LeftProperty },
    { "LEFT.TORP.LCOUNT",         ivsTorpLCount,    LeftProperty },
    { "LEFT.TORP.SHORT",          ivsTorpShort,     LeftProperty },
    { "LEFT.TYPE",                ivsType,          LeftProperty },
    { "LEFT.TYPE.SHORT",          ivsTypeShort,     LeftProperty },
    { "MAGIC",                    ivpMagic,         VcrProperty },
    { "RIGHT",                    ivsNameFull,      RightProperty },
    { "RIGHT.AUX",                ivsAuxName,       RightProperty },
    { "RIGHT.AUX$",               ivsAuxId,         RightProperty },
    { "RIGHT.AUX.AMMO",           ivsAuxAmmo,       RightProperty },
    { "RIGHT.AUX.COUNT",          ivsAuxCount,      RightProperty },
    { "RIGHT.AUX.SHORT",          ivsAuxShort,      RightProperty },
    { "RIGHT.BEAM",               ivsBeamName,      RightProperty },
    { "RIGHT.BEAM$",              ivsBeamId,        RightProperty },
    { "RIGHT.BEAM.COUNT",         ivsBeamCount,     RightProperty },
    { "RIGHT.BEAM.SHORT",         ivsBeamShort,     RightProperty },
    { "RIGHT.CREW",               ivsCrew,          RightProperty },
    { "RIGHT.CREW$",              ivsCrewRaw,       RightProperty },
    { "RIGHT.DAMAGE",             ivsDamage,        RightProperty },
    { "RIGHT.FIGHTER.BAYS",       ivsFighterBays,   RightProperty },
    { "RIGHT.FIGHTER.COUNT",      ivsFighterCount,  RightProperty },
    { "RIGHT.HULL",               ivsHullName,      RightProperty },
    { "RIGHT.HULL$",              ivsHullId,        RightProperty },
    { "RIGHT.ID",                 ivsId,            RightProperty },
    { "RIGHT.IMAGE",              ivsImage,         RightProperty },
    { "RIGHT.LEVEL",              ivsLevel,         RightProperty },
    { "RIGHT.MASS",               ivsMass,          RightProperty },
    { "RIGHT.NAME",               ivsName,          RightProperty },
    { "RIGHT.OWNER",              ivsOwnerShort,    RightProperty },
    { "RIGHT.OWNER$",             ivsOwnerId,       RightProperty },
    { "RIGHT.OWNER.ADJ",          ivsOwnerAdj,      RightProperty },
    { "RIGHT.SHIELD",             ivsShield,        RightProperty },
    { "RIGHT.STATUS",             ivsStatus,        RightProperty },
    { "RIGHT.TORP",               ivsTorpName,      RightProperty },
    { "RIGHT.TORP$",              ivsTorpId,        RightProperty },
    { "RIGHT.TORP.COUNT",         ivsTorpCount,     RightProperty },
    { "RIGHT.TORP.LCOUNT",        ivsTorpLCount,    RightProperty },
    { "RIGHT.TORP.SHORT",         ivsTorpShort,     RightProperty },
    { "RIGHT.TYPE",               ivsType,          RightProperty },
    { "RIGHT.TYPE.SHORT",         ivsTypeShort,     RightProperty },
    { "SEED",                     ivpSeed,          VcrProperty },
    { "TYPE$",                    ivpType,          VcrProperty },
};

IntVcrContext::IntVcrContext(int vcrnr)
    : vcrnr(vcrnr)
{ }

IntVcrContext::~IntVcrContext()
{ }

// IntContext:
int32_t
IntVcrContext::lookup(const string_t& name)
{
    return lookupName(name, vcr_mapping, countof(vcr_mapping));
}

void
IntVcrContext::set(int32_t /*index*/, IntValue* /*value*/)
{
    throw IntError::notAssignable();
}

IntValue*
IntVcrContext::get(int32_t index)
{
    /* Attempt to fetch VCR entry */
    if (!haveDisplayedTurn())
        return 0;

    GClassicVcrDatabase* cdb = dynamic_cast<GClassicVcrDatabase*>(&getDisplayedTurn().getVcrs());
    if (cdb == 0 || vcrnr > cdb->getCount())
        return 0;

    /* We have a VCR, and we have a valid index, so this will work out */
    GVcrEntry& vcr = cdb->getEntry(vcrnr-1);
    switch (vcr_mapping[index].domain) {
     case VcrProperty:
        return getVcrProperty(vcr, *cdb, IntVcrProperty(vcr_mapping[index].index));
     case LeftProperty:
        return getVcrSideProperty(vcr, 0, cdb, IntVcrSideProperty(vcr_mapping[index].index));
     case RightProperty:
        return getVcrSideProperty(vcr, 1, cdb, IntVcrSideProperty(vcr_mapping[index].index));
     default:
        return 0;
    }
}

bool
IntVcrContext::next()
{
    if (haveDisplayedTurn() && vcrnr < getDisplayedTurn().getVcrs().getCount()) {
        ++vcrnr;
        return true;
    } else {
        return false;
    }
}

IntContext*
IntVcrContext::clone()
{
    return new IntVcrContext(vcrnr);
}

GObject*
IntVcrContext::getObject()
{
    return 0;
}

string_t
IntVcrContext::toString(bool /*readable*/)
{
    return format("Vcr(%d)", vcrnr);
}

void
IntVcrContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notAssignable();
}

/****************************** Name Lookup ******************************/

/** Convert status flags to script value.
    \param status Status value
    \param side Side to query */
static IntValue*
makeVcrResult(int status, int side)
{
    switch (status) {
     case 0:
        return 0;
     case VcrPlayer::VCRS_INVALID:
        return makeStringValue("Invalid");
     case VcrPlayer::VCRS_TIMEOUT:
        return makeStringValue("Timeout");
     case VcrPlayer::VCRS_STALEMATE:
        return makeStringValue("Stalemate");
     case VcrPlayer::VCRS_LEFT_DESTROYED:
        return makeStringValue(side == 0 ? "Exploded" : "Survived");
     case VcrPlayer::VCRS_RIGHT_DESTROYED:
        return makeStringValue(side != 0 ? "Exploded" : "Survived");
     case VcrPlayer::VCRS_LEFT_CAPTURED:
        return makeStringValue(side == 0 ? "Captured" : "Survived");
     case VcrPlayer::VCRS_RIGHT_CAPTURED:
        return makeStringValue(side != 0 ? "Captured" : "Survived");
     case VcrPlayer::VCRS_RIGHT_DESTROYED + VcrPlayer::VCRS_LEFT_DESTROYED:
     default:
        return makeStringValue("Disabled");
    }
}

/** Get property of a VCR side. Note that the outcome is a property of
    a VCR side. Therefore, this function may need to invoke the VCR
    player. For this reason, the GVcrEntry cannot be const, because it
    will be updated with the computed result.

    The VCR can only be called through a GClassicVcrDatabase. When you
    do not pass in a GClassicVcrDatabase, the VCR will not be invoked,
    and the ivsStatus property will be empty for uncomputed VCRs. This
    can be used for the server module to make sure that a client who
    requests VCRs does not block our CPU. This will also guarantee
    that the GVcrEntry is not changed.

    \param vcr The record.
    \param side Side to query (0=left, 1=right)
    \param cdb Associated database, may be null
    \param ivs Property to get */
IntValue*
getVcrSideProperty(GVcrEntry& vcr, int side, GClassicVcrDatabase* cdb, IntVcrSideProperty ivs)
{
    /* Note: this implements the owner properties directly instead of relying on
       getPlayerProperty() like most others. This makes it possible to use VCR
       properties without having a turn loaded. */
    /* FIXME: this does not implement PlanetsHaveTubes; for planets with tubes,
       it always returns the fighters. */
    const TVcr& v = vcr.getVcr();
    const TVcrObject& o = v.objects[side];
    switch (ivs) {
     case ivsAuxAmmo:
        if (o.bay_count > 0) {
            return makeIntValue(v.num_fighters[side]);
        } else if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeIntValue(v.num_torps[side]);
        } else {
            return makeIntValue(0);
        }
     case ivsAuxCount:
        if (o.bay_count > 0) {
            return makeIntValue(o.bay_count);
        } else if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeIntValue(v.torp_launcher_count[side]);
        } else {
            return 0;
        }
     case ivsAuxId:
        if (o.bay_count > 0) {
            return makeIntValue(11);
        } else if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeIntValue(o.torp_type);
        } else {
            return 0;
        }
     case ivsAuxName:
        if (o.bay_count > 0) {
            return makeStringValue("Fighters");
        } else if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.torp_type).getName());
        } else {
            return 0;
        }
     case ivsAuxShort:
        if (o.bay_count > 0) {
            return makeStringValue("Ftr");
        } else if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.torp_type).getShortName());
        } else {
            return 0;
        }
     case ivsFighterBays:
        return makeIntValue(o.bay_count);
     case ivsFighterCount:
        return makeIntValue(o.bay_count > 0 ? v.num_fighters[side] : 0);
     case ivsTorpId:
        if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeIntValue(o.torp_type);
        } else {
            return 0;
        }
     case ivsTorpCount:
        if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeIntValue(v.num_torps[side]);
        } else {
            return makeIntValue(0);
        }
        
     case ivsTorpLCount:
        if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeIntValue(v.torp_launcher_count[side]);
        } else {
            return makeIntValue(0);
        }
     case ivsTorpShort:
        if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.torp_type).getShortName());
        } else {
            return 0;
        }
     case ivsTorpName:
        if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
            return makeStringValue(::getTorp(o.torp_type).getName());
        } else {
            return 0;
        }
     case ivsBeamCount:
        return makeIntValue(o.beam_count);
     case ivsBeamId:
        return makeIntValue(o.beam_type);
     case ivsBeamName:
        if (o.beam_type > 0 && o.beam_type <= NUM_BEAMS) {
            return makeStringValue(::getBeam(o.beam_type).getName());
        } else {
            return 0;
        }
     case ivsBeamShort:
        if (o.beam_type > 0 && o.beam_type <= NUM_BEAMS) {
            return makeStringValue(::getBeam(o.beam_type).getShortName());
        } else {
            return 0;
        }
     case ivsCrew:
        if (vcr.isPlanet(side)) {
            return 0;
        } else {
            return makeIntValue(o.crew);
        }
     case ivsCrewRaw:
        return makeIntValue(o.crew);
     case ivsDamage:
        return makeIntValue(o.damage);
     case ivsId:
        return makeIntValue(o.id);
     case ivsMass:
        return makeIntValue(v.mass[side]);
     case ivsName:
        return makeStringValue(convertGameToUtf8(getBasicString(o.name)));
     case ivsNameFull:
        return makeStringValue(format(vcr.isPlanet(side)
                                      ? _("%s (Planet #%d)")
                                      : _("%s (Ship #%d)"),
                                      convertGameToUtf8(getBasicString(o.name)),
                                      o.id));
     case ivsOwnerAdj:
        if (o.owner > 0 && o.owner <= NUM_OWNERS) {
            return makeStringValue(player_racenames.getAdjName(o.owner));
        } else {
            return 0;
        }
     case ivsOwnerId:
        return makeIntValue(o.owner);
     case ivsOwnerShort:
        if (o.owner > 0 && o.owner <= NUM_OWNERS) {
            return makeStringValue(player_racenames.getShortName(o.owner));
        } else {
            return 0;
        }
     case ivsShield:
        return makeIntValue(v.shield[side]);
     case ivsStatus:
        {
            /* If we have a status, return that */
            if (int st = vcr.getStatus())
                return makeVcrResult(st, side);

            /* We can only compute it when we have a database with some meta-information */
            if (!cdb)
                return 0;

            /* We can only compute it when we can get a player */
            Ptr<VcrPlayer> ply = cdb->createPlayer();
            if (!ply)
                return 0;
            ply->setVisualizer(new VcrNullVisualizer());

            /* Check whether player accepts it */
            TVcr vcr_tmp = v;
            if (!ply->setCapabilities(cdb->getCapabilities()) || ply->checkVcr(vcr_tmp))
                return 0;

            /* Compute it */
            ply->playVcr(vcr_tmp);
            vcr.setResultFromPlayer(*ply);

            /* Return result */
            return makeVcrResult(vcr.getStatus(), side);
        }
     case ivsType:
     case ivsTypeShort:
        {
            const char* kind;
            if (vcr.isPlanet(side)) {
                kind = "Planet";
            } else if (o.bay_count > 0) {
                kind = "Carrier";
            } else if (v.torp_launcher_count[side] > 0 && o.torp_type > 0 && o.torp_type <= NUM_TORPS) {
                kind = "Torpedo Ship";
            } else if (o.beam_count > 0) {
                kind = "Beam Weapons";
            } else {
                kind = "Freighter";
            }
            if (ivs == ivsTypeShort) {
                return makeStringValue(string_t(kind, 1));
            } else {
                return makeStringValue(kind);
            }
        }
        break;
     case ivsHullName:
     case ivsHullId:
        if (int h = vcr.getGuessedHull(side)) {
            if (ivs == ivsHullId) {
                return makeIntValue(h);
            } else {
                return makeStringValue(::getHull(h).getName());
            }
        } else {
            return 0;
        }
     case ivsImage:
        return makeIntValue(vcr.getShipPicture(side));
     case ivsLevel:
        return makeIntValue(o.experience_level);
    }
    return 0;
}

/** Get property of a VCR record.
    \param vcr The record
    \param cdb Associated database
    \param ivp Property to get */
IntValue*
getVcrProperty(const GVcrEntry& vcr, const GClassicVcrDatabase& cdb, IntVcrProperty ivp)
{
    switch (ivp) {
     case ivpSeed:
        return makeIntValue(vcr.getVcr().random_seed);
     case ivpMagic:
        return makeIntValue(vcr.getVcr().signature);
     case ivpType:
        return makeIntValue(vcr.getVcr().battle_type);
     case ivpAlgorithm:
        return makeStringValue(cdb.getTypeName(cdb.getType()));
     case ivpFlags:
        return makeIntValue(cdb.getCapabilities());
    }
    return 0;
}
