/**
  *  \file int/simple.cc
  *  \brief Interpreter: Simple Interface
  *
  *  This contains a few "one-call" interfaces for users to execute
  *  script actions.
  *
  *  (c) 2001-2011 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "int/simple.h"
#include "int/lex.h"
#include "int/interror.h"
#include "int/expr.h"
#include "int/parseexpr.h"
#include "int/exec.h"
#include "int/intglobal.h"
#include "int/csource.h"
#include "util/console.h"
#include "int/statement.h"
#include "io/streams.h"
#include "io/textfile.h"
#include "game/global.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "io/dirs.h"
#include "int/compound.h"
#include "util/except.h"
#include "int/process.h"

/** Compile expression, simple interface. Compiles the expression into a byte-code object.
    Returns the byte-code object, null on failure. This is used when an expression is used
    behind a regular UI function, where it doesn't really matter what kind of error, if any,
    we've encountered.

    \todo This would be a nice place to implement caching. */
IntBCORef
compileExpression(const string_t& expr, const IntCompilationContext& cc)
{
    try {
        IntTokenizer tok(expr);
        if (tok.getCurrentToken() == tok.tEnd)
            return 0; /* empty expression */

        std::auto_ptr<IntExprNode> expr(IntExpressionParser(tok).parse());
        if (tok.getCurrentToken() != tok.tEnd)
            return 0; /* expression incorrectly terminated */

        IntBCORef bco = new IntBytecodeObject();
        expr->compileValue(*bco, cc);
        bco->relocate();
        return bco;
    }
    catch (IntError& e) {
        return 0;
    }
}

/** Execute statement, simple interface. Compiles a one-line statement, using the current
    global context, and executes it. */
bool
executeStatement(const string_t& stmt, int prefixArg)
{
    // Create BCO
    IntBCORef bco = new IntBytecodeObject();

    // Create process
    IntExecutionContext& exec = createProcess(format(_("UI: %s"), stmt));
    exec.pushContextsFrom(global_contexts);
    exec.markContextTOS();

    // Remember prefix arg. FIXME: this stores the prefix arg in a global variable,
    // like in PCC 1.x. It would be cleaner to make a per-process static variable instead,
    // so that multiple processes can coexist.
    if (prefixArg > 0) {
        IntIntValue iv(prefixArg, false);
        exec.setVariable("UI.PREFIX", &iv);
    } else {
        exec.setVariable("UI.PREFIX", 0);
    }

    // Compile
    IntMemoryCommandSource mcs(stmt);
    IntStatementCompilationContext scc;
    scc.withExecutionContext(&exec);
    scc.withFlags(IntCompilationContext::RefuseBlocks);
    scc.withFlags(IntCompilationContext::LinearExecution);
    scc.withFlags(IntCompilationContext::ExpressionsAreStatements);

    bool result = true;
    try {
        IntStatementCompiler(mcs).compile(*bco, scc);
        bco->relocate();
        exec.pushNewFrame(new IntExecutionFrame(bco));
        runRunnableProcesses();
        if (exec.getState() == IntExecutionContext::psError) {
            const IntError& e = exec.getError();
            console.write(LOG_ERROR, e.what());
            string_t trace = e.getTrace();
            if (trace.size()) {
                console.write(LOG_INFO, trace, ConsoleRightJustified);
                result = false;
            }
        }
        killTerminatedProcesses();
    }
    catch (IntError& e) {
        console.write(LOG_ERROR, e.what());
        result = false;
    }

    // Update
    if (haveDisplayedTurn())
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();

    return result;
}

/** Execute a file, simple interface. */
bool
executeFile(const string_t& name, bool acceptMissingFile)
{
    bool result = true;
    try {
        // Open file
        Ptr<Stream> ps = Stream::openFile(name, Stream::C_READ);

        // Create process
        IntExecutionContext& exec = createProcess(format(_("File: %s"), name));
        exec.pushContextsFrom(global_contexts);
        exec.markContextTOS();

        // Generate compilation objects
        TextFile tf(*ps);
        IntFileCommandSource fcs(tf);
        IntBCORef nbco = new IntBytecodeObject();
        nbco->setFileName(ps->getName());

        // Compile
        try {
            IntStatementCompiler(fcs).compileList(*nbco,
                                                  IntStatementCompilationContext()
                                                  .withFlags(IntCompilationContext::LocalContext
                                                             | IntCompilationContext::ExpressionsAreStatements
                                                             | IntCompilationContext::LinearExecution));
        }
        catch (IntError& e) {
            e.addTrace(format(_("in file '%s', line %d"), Directory::getFileName(ps->getName()), fcs.getLineNumber()));
            throw;
        }

        nbco->relocate();

        exec.pushNewFrame(new IntExecutionFrame(nbco));
        runRunnableProcesses();
        killTerminatedProcesses();
    }
    catch (IntError& e) {
        // Script error
        console.write(LOG_ERROR, e.what(), ConsoleRightJustified);

        string_t trace = e.getTrace();
        if (trace.size())
            console.write(LOG_INFO, trace, ConsoleRightJustified);
        result = false;
    }
    catch (FileProblemException& e) {
        // File error
        if (!acceptMissingFile) {
            console.write(LOG_ERROR, e.what(), ConsoleRightJustified);
            result = false;
        }
    }

    // Update
    if (haveDisplayedTurn())
        getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
    return result;
}

/** Run a hook. Use this from non-script context. */
void
runHook(const string_t name)
{
    if (IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(global_values[global_property_names.lookup("ON " + name)])) {
        // This hook exists. Create a process and run it.
        IntExecutionContext& exec = createProcess(format(_("Hook: %s"), name));
        exec.pushContextsFrom(global_contexts);
        exec.markContextTOS();
        exec.pushNewFrame(new IntExecutionFrame(sv->getBytecodeObject()));
        runRunnableProcesses();
        killTerminatedProcesses();
    }
}
